## Copyright (c) 2016, Blake C. Rawlings.
##
## This file is part of `st2smv`.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

from ... import ast
from ... import ir
from ... import plugins
from ... import utils

logger = utils.logging.getLogger(__name__)

## Plugin information.
NAME = 'stdlib'
ACTIONS = (
    plugins.actions.SPECIAL_EXPRESSIONS,
    plugins.actions.UNTOUCHABLE,
)
OPTIONS = {}

DT = 'DT'
SPECIAL_EXPRESSION_TYPES = (DT,)


def convert_expression(converter, tree, lhs=False):
    """Convert the AST `tree` that corresponds to an expression.

    This modifies `self` and returns the converted expression.

    """
    assert tree[0] in SPECIAL_EXPRESSION_TYPES
    if tree[0] == 'DT':
        assert lhs is False
        return _convert_delay_timer(converter, tree)
    else:
        raise ir.ConversionError(
            'Unrecognized expression: {0}'.format(tree)
        )


def _convert_delay_timer(converter, tree):
    """Convert a call to `DT(reading, duration, default)` (a delay timer).

     This modifies `self`.

    """
    assert len(tree) == 2
    assert len(tree[1]) == 3
    assert tree[0] == 'DT'
    ## Check usage.
    if converter.if_depth > 0:
        raise ir.ConversionError(
            'Unsupported input code: DT inside IF...ELSE'
        )
    ## Handle the arguments to the `DT` function call; `duration`
    ## is ignored.
    reading = converter.convert_expression(tree[1][0])
    default = converter.convert_expression(tree[1][2])
    converter.set_type(reading, ast.BOOLEAN)
    ## Create the variable that will represent the timer.
    timer_variable = converter.convert_expression('dt', lhs=True)
    converter.assigned.remove(ir.Var('dt', None)) # only `dt.value` is assigned
    timer_value = ir.Var(
        timer_variable.label, timer_variable.version,
        field='value',
    )
    converter.referenced.add(timer_value)
    converter.defined.add(timer_value)
    _mark_timer(converter, timer_variable, timer_value)
    converter.set_type(timer_value, ast.BOOLEAN)
    ## Save the intermediate representation.
    converter.logic_ir.append(
        ['VAR', timer_variable, ['delay_timer', reading, default]]
    )
    ## Mark the connections.
    reading_symbols = ir.get_variables(reading)
    variables = set.union(converter.defined, converter.referenced)
    reading_vars = {x for x in variables if x in reading_symbols}
    for rvar in reading_vars:
        converter.mark_connection(timer_value, rvar)
    ## Return the expression that holds the timer's value.
    return timer_value


def _mark_timer(converter, timer, timer_value):
    if 'timers' not in converter.special_variables:
        converter.special_variables['timers'] = set()
    if 'timer_values' not in converter.special_variables:
        converter.special_variables['timer_values'] = set()
    converter.special_variables['timers'].add(timer)
    converter.special_variables['timer_values'].add(timer_value)


def is_untouchable(converter, variable):
    """Return whether or not `variable` is "untouchable".

    A variable is "untouchable" if it cannot (or should not) be
    abstracted away when simplifying the model.

    """
    timers = converter.special_variables.get('timers', set())
    timer_values = converter.special_variables.get('timer_values', set())
    if variable in (timers | timer_values):
        return True
    else:
        return False



## Copyright (c) 2016, Blake C. Rawlings.
##
## This file is part of `st2smv`.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import copy

from ... import ast
from ... import ir
from ... import plugins
from ... import smv
from ... import utils

logger = utils.logging.getLogger(__name__)

## Plugin information.
NAME = 'irrelevant_logic'
ACTIONS = (
    plugins.actions.PROCESS_INDEPENDENT_TESTS,
    plugins.actions.POSTPROCESS,
)
OPTIONS = {}


def generate_pits(converter):
    pit_tmp = converter.assigned | converter.defined
    pit_tmp -= converter.aliases
    pit_tmp -= converter.ivars
    pit_tmp &= converter.unconditional_assignments
    pit_vars = sorted(
        {x for x in pit_tmp if (converter.get_type(x) == ast.BOOLEAN)}
    )

    ##
    alternative_logic = {}
    locations = utils.get_locations(
        converter.logic_ir,
        rule=(lambda x: x == ast.ASSIGN)
    )
    assignments = [
        utils.get_value_at_location(converter.logic_ir, location[:-1])
        for location in locations
    ]
    for assignment in assignments:
        (lhs, rhs) = assignment[1]
        if lhs not in pit_vars:
            ## Don't need the alternative logic.
            continue
        logger.debug('Generating alternative logic for {}'.format(lhs))
        rhs_alternatives = get_rhs_alternatives(rhs)
        if rhs_alternatives is not None:
            assert lhs not in alternative_logic
            alternative_logic[lhs] = rhs_alternatives
    ##

    irrel = {'PIT_PREFIX': 'irrel'}
    spec_types = (
        ('SPEC', 'irrel_spec'),
        ('SYNTH', 'irrel_synth'),
    )
    for var in pit_vars:
        if var not in alternative_logic:
            continue
        assert var not in irrel
        irrel[var] = []
        for pattern in sorted(alternative_logic[var]):
            altvar = '{}_{}'.format(var, pattern)
            irrel[var].append('VAR {} : boolean;'.format(altvar))
            irrel[var].append('INIT {} = {};'.format(altvar, var))
            irrel[var].append(
                'TRANS initializing -> next({} = {});'.format(altvar, var)
            )
            irrel[var].append(
                'TRANS next({} = ({}));'
                .format(
                    altvar,
                    smv.ir_to_expression(alternative_logic[var][pattern])
                )
            )
            for spec_type in spec_types:
                irrel[var].append(
                    '{} NAME {}_{} := AG({} = {});'
                    .format(spec_type[0], spec_type[1], altvar, var, altvar)
                )
    return irrel


def get_rhs_alternatives(tree):
    """Compute "alternative versions" of the IR expression `tree`.

    This returns the `dict` of "alternative" expressions obtained by
    removing terms from `tree`.  The returned `dict` maps strings
    indicating which terms were dropped/kept (for example, the string
    '011' means the first term was dropped, and the other two were
    kept) to the corresponding expression.

    """
    def looks_like_boolean_expression(x):
        if isinstance(x, ir.Var):
            return True
        elif isinstance(x, (tuple, list)):
            if len(x) > 0:
                if x[0] in ast.comparators + (ir.EQ_NUMERIC, ir.NEQ_NUMERIC):
                    return True
        else:
            return False
    ## Generate the replacement `patterns`, based on the total number
    ## of `terms`.
    patterns = []
    locations = utils.get_locations(tree, rule=looks_like_boolean_expression)
    terms = [utils.get_value_at_location(tree, l) for l in locations]
    n = len(terms)
    if n <= 1:
        return None
    for i in range(n):
        pattern = [1] * n
        pattern[i] = 0
        patterns.append(tuple(pattern))

    ## Generate the alternative expressions to match the replacements
    ## `patterns`.
    alternatives = {}
    for pattern in patterns:
        key = ''.join((str(x) for x in pattern))
        alternatives[key] = _get_alternative_logic(
            tree, pattern, locations=locations
        )

    for alt in sorted(alternatives):
        logger.debug('{}: {}'.format(alt, alternatives[alt]))
    return alternatives


def _get_alternative_logic(tree, pattern, locations=None):
    """Delete some of the `terms` from `tree` according to `pattern`."""
    if locations is None:
        assert False
    else:
        terms = [utils.get_value_at_location(tree, l) for l in locations]
    assert len(pattern) == len(locations)

    ## Replace each dropped term with a dummy variable, `dropped`.
    dropped = ir.Var('DROPPED', None)
    while dropped in terms:
        dropped = ir.Var('_' + dropped.label, None)
    replaced = copy.deepcopy(tree)
    for i, p in enumerate(pattern):
        if p == 0:
            ## Drop the term.
            utils.set_value_at_location(replaced, locations[i], dropped)

    ## Reduce `tree`.
    reduced = _reduce_dropped_terms(replaced, dropped)
    return reduced


def _reduce_dropped_terms(tree, dropped):
    """Reduce `tree` by removing all appearances of `dropped`."""
    while True:
        reduced = _reduce_dropped_terms_worker(tree, dropped)
        if reduced == tree:
            break
        else:
            tree = reduced
    if dropped == reduced:
        logger.warning('Computed irrelevant logic with a single term.')
    return reduced


def _reduce_dropped_terms_worker(tree, dropped):
    """Reduce `tree` by removing an appearance of `dropped`."""
    ## Copy `tree` for modification (don't mess up the original).
    reduced = copy.deepcopy(tree)
    ## Find the first appearance of `dropped`.
    locations = utils.get_locations(tree, rule=(lambda x: x == dropped))
    if locations is None:
        return reduced
    if len(locations) == 0:
        return reduced
    location = locations[0]
    assert utils.get_value_at_location(tree, location) == dropped
    ## Find the location (and type) of the enclosing operation.
    op_location = location[:-1]
    while True:
        operation = utils.get_value_at_location(reduced, op_location)
        if operation[0] in ast.binary_operators + ast.unary_operators:
            operator = operation[0]
            op_distance = len(location) - len(op_location)
            break
        else:
            if len(op_location) == 0:
                assert False
            op_location = op_location[:-1]
    if op_distance == 1:
        assert operator in ast.unary_operators
    else:
        assert op_distance == 2
        assert operator in ast.binary_operators
    ## Remove `dropped` from the operation, and simplify.
    if operator in ast.unary_operators:
        if len(op_location) == 0:
            reduced = dropped
        else:
            utils.set_value_at_location(reduced, op_location, dropped)
    elif operator in ast.binary_operators:
        other_location = copy.copy(location)
        other_location[-1] = 1 - other_location[-1]
        new_operation = utils.get_value_at_location(reduced, other_location)
        if len(op_location) == 0:
            reduced = new_operation
        else:
            utils.set_value_at_location(
                reduced, op_location, new_operation
            )
    else:
        assert False
    ## Return the reduced IR tree.
    return reduced


def get_postprocessing_categories():
    return (
        ('Guaranteed Irrelevant Logic', 'irrel_spec', 'true'),
        ('Potential Irrelevant Logic', 'irrel_synth', 'true'),
        ('Guaranteed Relevant Logic', 'irrel_synth', 'false'),
        ('Potential Relevant Logic', 'irrel_spec', 'false'),
    )



## Copyright (c) 2015-2016, Blake C. Rawlings.
##
## This file is part of `st2smv`.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import copy

import networkx

from . import bfs
from ... import ast
from ... import plugins
from ... import utils

logger = utils.logging.getLogger(__name__)

## Plugin information.
NAME = 'connectivity'
ACTIONS = (
    plugins.actions.ABSTRACTION,
    plugins.actions.INFLUENCE,
)
OPTIONS = {
    'max_coi': (
        int,
        (
            'set the maximum cone-of-influence (COI) size, '
            'and compute COI cuts to enforce that limit'
        )
    )
}
MAX_COI_DEFAULT = 100


class Connectivity(object):
    """A class to analyze and manipulate the connections between the
    variables in a ST model.

    """
    def __init__(self, converter):
        ## Copy over the connection information.
        self.connections = copy.deepcopy(converter.connections)
        ## Set up a map from variables in `converter.connections` to a
        ## simpler type.
        self.var2node = self._generate_var2node_mapping()
        self.node2var = self._generate_node2var_mapping()
        if True:
            ## TODO: remove this test
            assert self.node2var == utils.invert_mapping(self.var2node)
        ## Produce a (di)graph from the connectivity information.
        self.digraph = self._connectivity_digraph()

    def _generate_var2node_mapping(self):
        """Map variables in `self.connections` to a simpler type (`int` for
        now).  This should help the performance.

        """
        mapping = {}
        mapped = set()
        n = 0
        for key in self.connections:
            if key in mapped:
                pass
            else:
                mapping[key] = n
                mapped.add(key)
                n += 1
            for connection in self.connections[key]:
                if connection in mapped:
                    pass
                else:
                    mapping[connection] = n
                    mapped.add(connection)
                    n += 1
        return mapping

    def _generate_node2var_mapping(self):
        """This has to be called after `self.var2node` is set up."""
        mapping = {}
        for var in self.var2node:
            node = self.var2node[var]
            mapping[node] = var
        return mapping

    def _connectivity_digraph(self):
        """Use the information in `self.connections` to generate a
        connectivity graph (a `networkx.DiGraph` object).

        """
        digraph = networkx.DiGraph()
        for to_var in self.connections:
            for from_var in self.connections[to_var]:
                digraph.add_edge(self.var2node[from_var], self.var2node[to_var])
        return digraph


@utils.timing(log_fun=logger.info, log_text='Computed COI cuts: {:.2f}s')
def compute_abstractions(converter):
    """Compute "COI cuts" to simplify the model.

    Returns a `dict` that maps each of the variables in `coi_vars` to
    COI cuts that enforce `max_coi` for that variable.

    """
    ## Mark some variables untouchable.
    untouchable = get_untouchable(converter)
    ## Get the set of variables that might need COI cuts.
    coi_variables = get_coi_variables(converter)
    ## Actually compute the COI cut information.
    coi_cuts = {}
    conn = Connectivity(converter) # TODO: avoid creating this twice
    n = len(coi_variables)
    alive_increment = max(1, int(n / 10))
    if 'coi' not in converter.special_variables:
        converter.special_variables['coi'] = {}
    for idx, variable in enumerate(coi_variables):
        if idx % alive_increment == 0:
            logger.info(
                'Computing COI cuts for variable {}/{}'.format(idx+1, n)
            )
        (deleted, size, size0, target, coi) = get_variable_cuts(
            conn.digraph, conn.node2var,
            [variable], converter.plugin_options, untouchable,
        )
        coi_cuts[variable] = {
            'deleted': sorted(deleted),
            'size': size,
            'original size': size0,
            'target': target,
            'coi': sorted(coi),
        }
        converter.special_variables['coi'][variable] = coi
    ## Return the COI cut information.
    return ('coi_cuts', coi_cuts)


def get_variable_cuts(
        digraph, node2var, coi_vars, plugin_options, untouchable, method=bfs
):
    """This is a very thin wrapper to compute COI cuts using `method`."""
    target = get_max_coi(plugin_options)
    (deleted, size, size0, coi) = method.get_variable_cuts(
        digraph, node2var,
        coi_vars=coi_vars,
        max_coi=target,
        untouchable=untouchable
    )
    return (deleted, size, size0, target, coi)


def get_max_coi(plugin_options):
    max_coi = plugin_options.get('max_coi')
    if max_coi is None:
        logger.warning(
            'Maximum COI size was not specified '
            '(with the "max_coi=n" plugin option), '
            'defaulting to {}'
            .format(MAX_COI_DEFAULT)
        )
        max_coi = MAX_COI_DEFAULT
    return max_coi


def get_untouchable(converter):
    """Get the set of variables that should not be abstracted."""
    untouchable = []
    might_be_untouchable = set.union(
        converter.referenced, converter.aliases, converter.trivial,
        converter.defined, converter.assigned
    )
    untouchable += [
        x for x in might_be_untouchable if converter.is_untouchable(x)
    ]
    for plugin in plugins.actions.get_active_plugins(
            converter.loaded_plugins, plugins.actions.UNTOUCHABLE
    ):
        untouchable += [
            x
            for x in might_be_untouchable
            if plugin.is_untouchable(converter, x)
        ]
        logger.debug(
            'Marked variables "untouchable" '
            'according to the "{}" plugin.'
            .format(plugin.NAME)
        )
    untouchable_not_boolean = {
        x
        for x in might_be_untouchable
        if (converter.get_type(x) != ast.BOOLEAN)
    }
    untouchable += list(untouchable_not_boolean)
    return untouchable


def get_coi_variables(converter):
    """Get the set of variables that might need COI cuts."""
    variables = converter.assigned | converter.defined
    numeric_variables = {
        x
        for x in variables
        if (converter.get_type(x) == ast.NUMERIC)
    }
    if True:
        ## TODO: remove this check.
        numeric_variables_old = {
            x
            for x in variables
            if (converter.is_numeric(x) or converter.is_numeric(x.label))
        }
        assert numeric_variables == numeric_variables_old
    return variables - numeric_variables


def get_coi_info(converter):
    """Get the info that is needed to compute COI cuts."""
    conn = Connectivity(converter) # TODO: avoid creating this twice
    untouchable = set()
    for variable in get_untouchable(converter):
        node = conn.var2node.get(variable)
        if node is not None:
            untouchable.add(node)
    return {
        'edges': conn.digraph.edges(),
        'node2var': conn.node2var,
        'untouchable': sorted(untouchable),
    }


def get_cone_of_influence(converter, variable):
    return converter.special_variables['coi'][variable]



## Copyright (c) 2015-2016, Blake C. Rawlings.
##
## This file is part of `st2smv`.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import sys
import time

import pyparsing as p

from . import ast
from . import utils

logger = utils.logging.getLogger(__name__)

## Increase the maximum recursion depth to parse deeply-nested
## expressions.
sys.setrecursionlimit(2000)

#############
# Operators #
#############

## Assignment.
COLEQ = p.Literal(':=').setParseAction(lambda x: ast.ASSIGN)

## Function block call arguments.
FBARG = p.Literal(':=').setParseAction(lambda x: ast.FBARG)
FBOUT = p.Literal('=>').setParseAction(lambda x: ast.FBOUT)

## Comparison.
LT = p.Literal('<').setParseAction(lambda x: ast.LT)
LEQ = p.Literal('<=').setParseAction(lambda x: ast.LEQ)
GT = p.Literal('>').setParseAction(lambda x: ast.GT)
GEQ = p.Literal('>=').setParseAction(lambda x: ast.GEQ)
NEQ = p.Literal('<>').setParseAction(lambda x: ast.NEQ)
EQ = p.Literal('=').setParseAction(lambda x: ast.EQ)
## Order is important here!  If it was 'LT | LEQ', then when matching
## an expression with '<=', this would match the '<' first, then hit
## the '=', and fail.
binop_comp = (NEQ | LEQ | LT | GEQ | GT | EQ)

## Boolean.
AND = p.CaselessKeyword('AND').setParseAction(lambda x: ast.AND)
OR = p.CaselessKeyword('OR').setParseAction(lambda x: ast.OR)
XOR = p.CaselessKeyword('XOR').setParseAction(lambda x: ast.XOR)
NOT = p.CaselessKeyword('NOT').setParseAction(lambda x: ast.NOT)
##
binop_bool = (AND | OR | XOR)
unop_bool = NOT

## Arithmetic.
ADD = p.Literal('+').setParseAction(lambda x: ast.ADD)
SUB = p.Literal('-').setParseAction(lambda x: ast.SUB)
MLT = p.Literal('*').setParseAction(lambda x: ast.MLT)
DIV = p.Literal('/').setParseAction(lambda x: ast.DIV)
EXP = p.Literal('**').setParseAction(lambda x: ast.EXP)
MOD = p.CaselessKeyword('MOD').setParseAction(lambda x: ast.MOD)
NEG = p.Literal('-').setParseAction(lambda x: ast.NEG)
##
binop_num = (EXP | MLT | SUB | ADD | DIV | MOD) # order: `EXP` before `MLT`
unop_num = NEG

## Convenience variables.
binop = (binop_comp | binop_bool | binop_num)
unop = (unop_bool | unop_num)

########################
# Symbols and Keywords #
########################

## Use .suppress() to prevent the symbol itself from showing up in the
## parsed output (e.g., if its meaning can be inferred from some other
## structure).
SEMICOLON = p.Literal(';').suppress()
COMMA = p.Literal(',').suppress()
LPAR = p.Literal('(').suppress()
RPAR = p.Literal(')').suppress()
TRUE = p.CaselessKeyword('TRUE').setParseAction(lambda x: ast.TRUE)
FALSE = p.CaselessKeyword('FALSE').setParseAction(lambda x: ast.FALSE)
IF = p.CaselessKeyword('IF').setParseAction(lambda x: ast.IF)
THEN = p.CaselessKeyword('THEN').suppress()
ELSIF = p.CaselessKeyword('ELSIF')
ELSE = p.CaselessKeyword('ELSE')
ENDIF = p.CaselessKeyword('END_IF').suppress()

##########
# Values #
##########

## Identifiers.
identifier = p.Word(p.alphas, p.alphanums + '_')
## Literals.
literal_bool = (TRUE | FALSE)
number_sequence = p.Word(p.nums)
literal_int = number_sequence
literal_real = p.Combine(
    (
        (p.Optional(number_sequence) + '.' + number_sequence)
        | (number_sequence + '.' + p.Optional(number_sequence))
    )
    + p.Optional(p.CaselessLiteral('E') + number_sequence)
)
literal_num = (literal_real | literal_int)
literal = (literal_bool | literal_num)
##
value = (literal | identifier)

############
# Comments #
############

comment_block = p.nestedExpr(opener='(*', closer='*)')
comment_line = p.Group('//' + p.restOfLine)
##
comment = (comment_block | comment_line)

###############
# Expressions #
###############

expression = p.Forward()
## Function calls.
grouping = p.Group(LPAR + expression + RPAR)
function_call = (
    identifier
    + LPAR
    + p.Optional(p.delimitedList(expression))
    + RPAR
)
function_call.setParseAction(
    lambda tokens: [[ast.CALL, [tokens[0], tokens[1:]]]]
)
## Function block calls.
_function_block_argument = (
    p.Group(identifier + FBARG + expression)
    | p.Group(identifier + FBOUT + identifier)
    | expression
)
function_block_call = (
    identifier
    + LPAR
    + p.Optional(p.delimitedList(_function_block_argument))
    + RPAR
)
function_block_call.setParseAction(
    lambda tokens: [[ast.FUNCTION_BLOCK_CALL, [tokens[0], tokens[1:]]]]
)
## Operator precedence via precedence levels (slow, lots of
## recursion!).
precedence_1 = p.Forward()
precedence_2 = p.Forward()
precedence_3 = p.Forward()
precedence_4 = p.Forward()
precedence_5 = p.Forward()
precedence_6 = p.Forward()
precedence_7 = p.Forward()
precedence_8 = p.Forward()
precedence_9 = p.Forward()
precedence_10 = p.Forward()
precedence_1 << (
    p.Group(precedence_2 + p.OneOrMore(OR + precedence_2))
    | precedence_2
)
precedence_2 << (
    p.Group(precedence_3 + p.OneOrMore(XOR + precedence_3))
    | precedence_3
)
precedence_3 << (
    p.Group(precedence_4 + p.OneOrMore(AND + precedence_4))
    | precedence_4
)
precedence_4 << (
    p.Group(precedence_5 + p.OneOrMore((EQ|NEQ) + precedence_5))
    | precedence_5
)
precedence_5 << (
    p.Group(precedence_6 + p.OneOrMore((LEQ|LT|GEQ|GT) + precedence_6))
    | precedence_6
)
precedence_6 << (
    p.Group(precedence_7 + p.OneOrMore((ADD|SUB) + precedence_7))
    | precedence_7
)
precedence_7 << (
    p.Group(precedence_8 + p.OneOrMore((MLT|DIV|MOD) + precedence_8))
    | precedence_8
)
precedence_8 << (
    p.Group((NEG|NOT) + precedence_8)
    | precedence_9
)
precedence_9 << (
    p.Group(precedence_10 + p.OneOrMore(EXP + precedence_10))
    | precedence_10
)
precedence_10 << (grouping | function_call | function_block_call | value)
## (Not used currently) operator precedence via a built-in `pyparsing`
## method (/incredibly/ slow without `.enablePackrat()`).
operation = p.infixNotation(
    (function_call | function_block_call | value),
    [
        (EXP, 2, p.opAssoc.LEFT),
        ((NEG|NOT), 1, p.opAssoc.RIGHT),
        ((MLT|DIV|MOD), 2, p.opAssoc.LEFT),
        ((ADD|SUB), 2, p.opAssoc.LEFT),
        ((LEQ|LT|GEQ|GT), 2, p.opAssoc.LEFT),
        ((EQ|NEQ), 2, p.opAssoc.LEFT),
        (AND, 2, p.opAssoc.LEFT),
        (XOR, 2, p.opAssoc.LEFT),
        (OR, 2, p.opAssoc.LEFT),
    ]
)
operation.enablePackrat()
##
expression << precedence_1

################
# Instructions #
################

instruction = p.Group((function_call | function_block_call) + SEMICOLON)

##############
# Statements #
##############

statement = p.Forward()
## Assignments.
assignment = (
    identifier + COLEQ + expression + SEMICOLON
)
## IF/ELSIF/ELSE.
elsif_block = (
    ELSIF + expression + THEN +
    p.Group(p.ZeroOrMore(statement))
)
else_block = (
    ELSE +
    p.Group(p.ZeroOrMore(statement))
)
if_elsif_else = (
    IF + expression + THEN +
    p.Group(p.ZeroOrMore(statement)) +
    p.Optional(p.OneOrMore(elsif_block)) +
    p.Optional(else_block) +
    ENDIF + SEMICOLON
)
##
statement << p.Group(
    if_elsif_else
    | assignment
    | instruction
)

###################
# String Literals #
###################

quoted_string = p.QuotedString('"', multiline=True)

###############
# Blank Lines #
###############

blank_line = SEMICOLON

#################
# The Main Loop #
#################

main_loop = p.OneOrMore(
    statement | quoted_string.suppress() | blank_line.suppress()
)
main_loop.ignore(comment)


@utils.timing(log_fun=logger.info, log_text='Parsed the ST file: {:.2f}s')
def _parse(path):
    """Parse the Structured Text file located at `path`."""
    with open(path) as f:
        lines = f.readlines()
    n_lines = len(lines)
    step_size = 1
    out = None
    fromline = 0
    toline = min(fromline + step_size, n_lines)
    while True:
        if fromline >= n_lines:
            ## Done parsing.  Strict inequality should never hold.
            break
        logger.debug(
            'Parsing lines {}-{} ({} lines) of {}'
            .format(fromline+1, toline, toline - fromline, n_lines)
        )
        working = '\n'.join(lines[fromline : toline])
        try:
            ## Parse and record.
            chunk = main_loop.parseString(working, parseAll=True)
            if out is None:
                out = chunk
            else:
                out += chunk
            ## Update for the next iteration.
            fromline = min(toline, n_lines)
            toline = min(fromline + step_size, n_lines)
        except p.ParseException:
            if toline == n_lines:
                ## Failed.
                break
            toline = min(toline + 1, n_lines)
            continue
    if fromline < n_lines:
        logger.warning(
            'Did not completely parse the file; failed after line {}'
            .format(fromline+1)
        )
    return out


def parse_symbol(s, expr_type):
    """Try to parse a single symbol of structured text code as a
    particular type of expression.

    """
    s = str(s)
    try:
        out = expr_type.parseString(s)[0]
    except p.ParseException:
        out = None
    return out


def pt_to_ast(tree):
    """Convert the parse tree to an abstract syntax tree (AST)."""
    ## Do some simple preprocessing.
    assert isinstance(tree, (p.ParseResults, list, utils.six.string_types))
    ##
    if isinstance(tree, (str, utils.six.string_types)):
        ## Base case, got to a symbol.
        return tree
    elif len(tree) == 0:
        return tree
    elif len(tree) == 1:
        ## A node with a single child, replace node with child.
        return pt_to_ast(tree[0])
    elif tree[0] in ast.unary_operators:
        ## Unary operator.
        assert len(tree) == 2
        return [tree[0], pt_to_ast(tree[1])]
    elif tree[0] == ast.IF:
        ## IF statement.
        assert len(tree) >= 3
        ## Check for ELSIF, replace with ELSE IF
        if ELSIF in tree[1:]:
            assert len(tree) >= 6
            assert tree[3] == parse_symbol(tree[3], ELSIF)
            return [
                tree[0],
                [
                    pt_to_ast(tree[1]),
                    pt_to_ast(tree[2]),
                    pt_to_ast([ast.IF] + tree[4:]),
                ],
            ]
        elif ELSE in tree[1:]:
            ## IF-ELSE statement.
            return [
                tree[0],
                [
                    pt_to_ast(tree[1]),
                    pt_to_ast(tree[2:-2]),
                    pt_to_ast(tree[-1]),
                ],
            ]
        else:
            ## Simple IF statement.
            assert len(tree) == 3
            return [
                tree[0],
                [
                    pt_to_ast(tree[1]),
                    pt_to_ast(tree[2]),
                    [],
                ],
            ]
    elif tree[0] in (ast.CALL, ast.FUNCTION_BLOCK_CALL):
        assert len(tree) == 2
        assert len(tree[1]) == 2
        return [tree[0], [tree[1][0], [pt_to_ast(x) for x in tree[1][1]]]]
    elif tree[1] == ast.ASSIGN:
        ## Assignment.
        assert len(tree) == 3
        return [tree[1], [pt_to_ast(tree[0]), pt_to_ast(tree[2])]]
    elif tree[1] in ast.function_block_operators:
        ## Function block argument "declaration".
        return [tree[1], [pt_to_ast(tree[0]), pt_to_ast(tree[2])]]
    elif tree[1] in (ast.binary_operators + ast.comparators):
        ## Binary operator.
        n = len(tree)
        assert n >= 3 and (n - 3) % 2 == 0
        if True:
            ## Left-associative.
            head = tree[0:-2]
            operator = tree[-2]
            assert operator in (ast.binary_operators + ast.comparators)
            tail = tree[-1]
        else:
            ## Right-associative.
            head = tree[0]
            operator = tree[1]
            tail = tree[2:]
        return [operator, [pt_to_ast(head), pt_to_ast(tail)]]
    else:
        ## Descend.
        return [pt_to_ast(x) for x in tree]


def get_ast(path):
    parse_tree = _parse(path)
    t0 = time.time()
    abstract_syntax_tree = pt_to_ast(parse_tree)
    t1 = time.time()
    logger.info('Created AST: {0:.2f}s'.format(t1 - t0))
    return abstract_syntax_tree



## Copyright (c) 2016, Blake C. Rawlings.
##
## This file is part of `st2smv`.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import copy
import itertools
import json
import os
import pprint

from . import plugins
from . import utils

logger = utils.logging.getLogger(__name__)

## Assignment.
ASSIGN = 'ASSIGN'

## Calls.
CALL = 'CALL'
FUNCTION_BLOCK_CALL = 'FUNCTION_BLOCK_CALL'

## Function block call arguments.
FBARG = 'FBARG'
FBOUT = 'FBOUT'
function_block_operators = (FBARG, FBOUT)

## Comparison.
LT = '<'
LEQ = '<='
GT = '>'
GEQ = '>='
NEQ = '<>'
EQ = '='
inequality_comparators = (LT, LEQ, GT, GEQ)
equality_comparators = (NEQ, EQ)
comparators = inequality_comparators + equality_comparators

## Boolean operators.
AND = 'AND'
OR = 'OR'
XOR = 'XOR'
NOT = 'NOT'
binary_boolean_operators = (AND, OR, XOR)
unary_boolean_operators = (NOT,)
boolean_operators = binary_boolean_operators + unary_boolean_operators

## Arithmetic operators.
ADD = '+'
SUB = '-'
MLT = '*'
DIV = '/'
EXP = '**'
MOD = 'MOD'
NEG = 'NEG'
binary_arithmetic_operators = (ADD, SUB, MLT, DIV, EXP, MOD)
unary_arithmetic_operators = (NEG,)
arithmetic_operators = binary_arithmetic_operators + unary_arithmetic_operators

## Higher level "operators".
CONJUNCTION = 'CONJUNCTION'
DISJUNCTION = 'DISJUNCTION'

## Values.
TRUE = 'TRUE'
FALSE = 'FALSE'

## Control flow.
IF = 'IF'

## Convenience variables.
binary_operators = binary_boolean_operators + binary_arithmetic_operators
unary_operators = unary_boolean_operators + unary_arithmetic_operators

## Types.
BOOLEAN = 'BOOLEAN'
NUMERIC = 'NUMERIC'
VALUE = 'VALUE'
NIL = 'NIL'


class AstError(Exception):
    """Error with the structure of the AST."""
    pass


def abs_to_disjunction(tree, abs_locs=None, recursive_call=False):
    """Given the abstract syntax tree (AST) `tree` of a numerical
    comparison/inequality predicate, and the list of locations in
    `tree` where absolute value forms appear, `abs_locs`, return the
    AST with no absolute value forms that represents the conjunction
    of disjunctions that is equivalent to `tree`.

    This should correctly handle strict and non-strict "greater than"
    and "less than" predicates, with absolute values on either side.
    It does *not* currently handle negated absolute value forms.

    """
    if not recursive_call:
        logger.debug('Top-level call to `abs_to_disjunction`.')
        assert abs_locs is None
    if abs_locs is None:
        abs_locs = find_absolute_value_locations(tree)
    n = len(abs_locs)
    if len(abs_locs) == 0:
        logger.debug('AST has no absolute values: {}'.format(tree))
        return tree

    logger.debug('AST with absolute values: {}'.format(tree))

    if tree[0] in ('<', '<='):
        lhs_test = lambda x: x[1] == 0
    elif tree[0] in ('>', '>='):
        lhs_test = lambda x: x[1] == 1
    elif tree[0] == '=':
        lhs_test = lambda x: False
    elif tree[0] == '<>':
        lhs_test = lambda x: True
    else:
        ## Not handling '=' or '<>' right now.
        assert False
    lhs_locs = []
    rhs_locs = []
    for loc in abs_locs:
        assert loc[0] == 1
        if lhs_test(loc):
            lhs_locs.append(loc)
        else:
            rhs_locs.append(loc)
    logger.debug('Left-hand-side absolute values at: {}'.format(lhs_locs))
    logger.debug('Right-hand-side absolute values at: {}'.format(rhs_locs))

    ## If any left-hand-side absolute values exist, then replace them;
    ## otherwise, replace the right-hand-side absolute values.
    if len(lhs_locs) > 0:
        n = len(lhs_locs)
        result_type = CONJUNCTION
        locations = lhs_locs
    else:
        n = len(rhs_locs)
        result_type = DISJUNCTION
        locations = rhs_locs
    assert n > 0
    patterns = list(itertools.product(*([[True, False]] * n)))

    transformed = []
    for pattern in patterns:
        variation = copy.deepcopy(tree)
        for i, p in enumerate(pattern):
            ## Warning/hack: modify the desired element of `variation`
            ## by reference here.
            element = variation
            for idx in locations[i][:-1]:
                element = element[idx]
            abs_form = element[locations[i][-1]]
            if p:
                ## Just take the argument of the absolute value (the
                ## "positive" version).
                abs_form[0] = 'POS'
            else:
                ## Negate the argument of the absolute value.
                abs_form[0] = NEG
        if result_type == CONJUNCTION:
            ## Make a recursive call to take care of the RHS.
            variation = abs_to_disjunction(
                variation, abs_locs=rhs_locs, recursive_call=True
            )
        else:
            pass
        transformed.append(variation)

    transformed = [result_type, transformed]
    logger.debug('Transformed AST: {}'.format(transformed))
    return transformed


def find_absolute_value_locations(tree, root=None):
    """Given the AST `tree`, return the list of (lists of) indices that
    correspond to absolute values.

    """
    if root is None:
        ## TODO: make sure this is a comparison.
        root = []
    if not isinstance(tree, (tuple, list)):
        return None
    elif len(tree) == 0:
        return None
    elif tree[0] == 'ABS':
        return [root]
    else:
        retval = []
        for idx, subtree in enumerate(tree):
            subroot = root + [idx]
            subretval = find_absolute_value_locations(subtree, root=subroot)
            if subretval is not None:
                retval += subretval
        return retval


def dump_ast(fname, parser, loaded_plugins=None):
    """Dump the AST of the file `fname` after parsing it with `parser`."""
    if loaded_plugins is None:
        loaded_plugins = []
    tree = parser.get_ast(fname)
    pprint.pprint(tree)
    ## Check for any plugins that do AST transformations, and apply
    ## them.
    for plugin in plugins.actions.get_active_plugins(
            loaded_plugins, plugins.actions.AST_TRANSORMATIONS
    ):
        transformer = plugin.Transformer()
        transformed = transformer.transform_ast(tree)
        tree = transformed
        logger.debug(
            'Applied AST transformations from the "{}" plugin.'
            .format(plugin.NAME)
        )
        print('### Transformed AST (via "{}" plugin): ###'.format(plugin.NAME))
        pprint.pprint(tree)


def write_ast(tree, directory=None):
    """Write the AST `tree` to a file in the specified `directory`."""
    if directory is None:
        pprint.pprint(tree)
    else:
        if not os.path.isdir(directory):
            os.mkdir(directory)
        path = directory + os.sep + 'ast.json'
        with open(path, 'w') as f:
            json.dump(tree, f)
            f.write('\n')



## Copyright (c) 2015-2016, Blake C. Rawlings.
##
## This file is part of `st2smv`.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import logging
import time

import six

logger = logging.getLogger(__name__)


def timing(log_fun=logger.debug, log_text='{}'):
    """Use this function as a decorator to log timing information.

    `log_text` should be a format string with a single placeholder,
    which will be replaced with the time it took to call the function.

    """
    def decorator(f):
        def wrapped(*args, **kwargs):
            t0 = time.time()
            retval = f(*args, **kwargs)
            t1 = time.time()
            log_fun(log_text.format(t1 - t0))
            return retval
        return wrapped
    return decorator


def try_incr(d, key, delta):
    """Try to increment `d[key]` by `delta`.

    If `key` is not in `d`, then add it to `d` with the value `delta`.

    """
    if key in d:
        d[key] += delta
    else:
        d[key] = delta


def flatten(nested):
    """Flatten a nested list."""
    ## Base case.
    if not isinstance(nested, (list, tuple)):
        return [nested]
    ## Recursive case.
    flat = []
    for n in nested:
        f = flatten(n)
        if isinstance(f, (list, tuple)):
            flat += f
        else:
            flat.append(f)
    return flat


def get_locations(tree, rule=None, root=None):
    """Get the (list of) locations in `tree` that meet `rule`.

    `tree` should have a nested iterable (`tuple` or `list`)
    structure.

    """
    if rule is None:
        rule = lambda x: True
    if root is None:
        root = []
    ##
    if rule(tree):
        return [root]
    elif isinstance(tree, (tuple, list)):
        retval = []
        for idx, subtree in enumerate(tree):
            subroot = root + [idx]
            subretval = get_locations(subtree, rule=rule, root=subroot)
            if subretval is not None:
                retval += subretval
        return retval
    else:
        return None


def get_value_at_location(tree, location):
    reference = tree
    for idx in location:
        reference = reference[idx]
    return reference


def set_value_at_location(tree, location, value):
    reference = tree
    max_depth = len(location)
    for depth, idx in enumerate(location):
        if depth < max_depth - 1:
            reference = reference[idx]
        else:
            reference[idx] = value


def invert_mapping(mapping):
    """Invert a mapping.

    The input should be a `dict` with a 1:1 mapping from key to value.

    """
    return {value: key for (key, value) in mapping.items()}



## Copyright (c) 2015-2016, Blake C. Rawlings.
##
## This file is part of `st2smv`.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import copy
import math

import networkx

from . import cone_of_influence
from ... import utils

logger = utils.logging.getLogger(__name__)


def disconnect_digraph(
        dg, limit,
        root=None, untouchable=None,
):
    """

    """
    ## Check the arguments.
    assert isinstance(root, list)
    assert isinstance(untouchable, list)

    ## Compute shortest paths and node depths.
    sp = _shortest_path_length_to_multiple_targets(dg, targets=root)
    if len(sp) <= limit:
        return set()
    depths = {}
    for node in sp:
        depth = sp[node]
        if depth not in depths:
            depths[depth] = set()
        depths[depth].add(node)
    assert depths == depths

    ## Get the target depth.
    target_depth = 0
    total = 0
    while True:
        total += len(depths[target_depth])
        if total >= min(limit, len(sp)):
            break
        else:
            target_depth += 1
            continue

    ## Delete any nodes `deeper` than the `target_depth`.
    deeper = set()
    for depth in range(target_depth + 1, max(depths) + 1):
        deeper |= depths[depth]
    also_deleted = deeper - set(untouchable)
    if len(also_deleted) > 0:
        return also_deleted

    ## Find the "candidate" nodes at or above the `target_depth`.
    candidate_depth = target_depth
    while True:
        if candidate_depth < 0:
            raise cone_of_influence.AbstractionError(
                'Cannot remove nodes below depth 0 (tried {}).'
                .format(candidate_depth)
            )
        candidates = depths[candidate_depth] - set(untouchable)
        if len(candidates) > 0:
            break
        else:
            candidate_depth -= 1
            continue

    ## Delete nodes to disconnect the graph.
    if len(candidates) == 0:
        deleted = set()
    else:
        ## Only delete the "major contributors".
        candidates_with_scores = sorted(
            [
                (x, _candidate_score(dg, root, x, depths))
                for x in candidates
            ],
            key=(lambda c_w_s: c_w_s[1]),
            reverse=True
        )
        deleted = {candidates_with_scores[0][0]}
    ##
    return deleted


def _candidate_score(dg, root, node, depths):
    ## Find the shortest path length from `root` to `node`.
    depth = _shortest_path_length_to_multiple_targets(
        dg, source=node, targets=root
    )

    ## Get the set of variables that are already included in the COI,
    ## up to `depth`.
    already_included = set()
    for d in sorted(depths):
        if d <= depth:
            already_included |= depths[d]
        else:
            break

    ## Get `node`'s additional contribution to the COI.
    node_predecessors = dg.predecessors(node)
    also_included = set(node_predecessors) - already_included

    ## Compute the cost.
    cost = len(also_included)
    return cost


def _shortest_path_length_to_multiple_targets(graph, source=None, targets=None):
    """Wrap `networkx.shortest_path_length` and add the ability to specify
    `targets` as a list of nodes (instead of a single node).

    """
    if isinstance(targets, list):
        ## Multiple targets.
        results = []
        for target in targets:
            try:
                result = networkx.shortest_path_length(
                    graph, source=source, target=target
                )
                results.append(result)
            except networkx.exception.NetworkXNoPath:
                ## `source` is a single node, and it does not have a
                ## path to this `target`; skip it.
                pass
        if source is None:
            ## No source; `results` will contain `dict`s.
            return _merge_length_dicts(results)
        else:
            ## Assume a single source; `results` will contain `int`s.
            return min(results)
    else:
        return networkx.shortest_path_length(
            graph, source=source, target=targets
        )


def _merge_length_dicts(length_dicts):
    assert isinstance(length_dicts, (list, tuple))
    lengths = dict()
    for ld in length_dicts:
        for source_node in ld:
            incumbent = lengths.get(source_node)
            challenger = ld[source_node]
            if incumbent is None:
                lengths[source_node] = challenger
            elif challenger < incumbent:
                lengths[source_node] = challenger
    return lengths


def cut_coi(
        dg, limit,
        root=None, untouchable=None,
):
    """

    """
    ## Check the arguments.
    assert isinstance(root, list)

    ## Make sure not to delete any `root` variables.
    if untouchable is None:
        untouchable = root[:]
    else:
        untouchable = untouchable + root[:]

    ## If the largest COI is larger than `limit`, then delete some
    ## nodes to bring the size of that COI down to `limit`.
    deleted = disconnect_digraph(
        dg, limit,
        root=root, untouchable=untouchable,
    )

    ## Return the set of deleted nodes.
    return deleted


def cut_coi_iterative(
        digraph, limit,
        root=None, untouchable=None,
        already_deleted=None
):
    """

    """
    ## Check the arguments.
    assert isinstance(root, list)
    if already_deleted is None:
        already_deleted = set()

    ## Work with a subgraph.
    coi_nodes = cone_of_influence.digraph_coi(digraph, root)
    dg = digraph.subgraph(coi_nodes)

    ## Remove the `already_deleted` nodes, and recompute the subgraph.
    for ad in already_deleted:
        successors = dg.successors(ad)
        dg.remove_node(ad)
        dg.add_node(ad)
        for s in successors:
            dg.add_edge(ad, s)
    coi_nodes = cone_of_influence.digraph_coi(dg, root)
    dg = dg.subgraph(coi_nodes)

    ## Mark any "leaf" variables in the COI that don't have connection
    ## information (probably input variables in the model) as
    ## untouchable, since deleting them would have no impact.
    u = [x for x in coi_nodes if dg.in_degree(x) == 0]
    if untouchable is None:
        pass
    else:
        u += copy.deepcopy(untouchable)

    ## Iteratively delete nodes and check if the COI limit is met.
    deleted = set()
    iterations = 0
    while True:
        try:
            new_deleted = cut_coi(
                dg, limit,
                root=root, untouchable=(u + list(deleted)),
            )
        except cone_of_influence.AbstractionError:
            new_deleted = set()
        if len(new_deleted) > 0:
            ## Replace the "deleted" nodes.
            for nd in new_deleted:
                successors = dg.successors(nd)
                dg.remove_node(nd)
                dg.add_node(nd)
                for s in successors:
                    dg.add_edge(nd, s)
            ## Update the subgraph.
            coi_nodes = cone_of_influence.digraph_coi(dg, root)
            dg = dg.subgraph(coi_nodes)
            ## Get ready for the next iteration.
            deleted = set.union(deleted, new_deleted)
            iterations += 1
            continue
        else:
            logger.debug('Did not delete any more nodes.')
            break

    ## Get the final COI information, then add the edges back.
    coi = cone_of_influence.digraph_coi(dg, root)

    ## Done.
    logger.debug(
        'After {} iteration(s), deleted {} node(s): {}'
        .format(iterations, len(deleted), deleted)
    )
    return (deleted | already_deleted, coi, iterations)


def get_variable_cuts(
        digraph, node2var,
        coi_vars=None,
        max_coi=100,
        untouchable=None,
):
    """Given the COI graph `digraph`, compute the set of variables that
    should be removed so that the COI rooted at `coi_vars` contains no
    more than `max_coi` variables.  Any variables in `untouchable`
    will not be removed.

    """
    logger.debug('Computing COI cuts for variable(s): {}'.format(coi_vars))

    ## Convert from "variables" (`ir.Var`) to "nodes" (`int`), which
    ## are used for all internal computations.
    var2node = utils.invert_mapping(node2var)
    if coi_vars is None:
        coi_nodes = None
    else:
        coi_nodes = [var2node[x] for x in coi_vars]

    ## Original size.
    size0 = -1
    if coi_nodes is not None:
        coi0 = cone_of_influence.digraph_coi(digraph, coi_nodes)
        size0 = len(coi0)

    ## Cuts.
    assert isinstance(max_coi, int)
    undershoot_tolerance = 0.8
    overshoot_tolerance = 1.0
    deleted = set()
    (deleted, coi, coi_iterations) = cut_coi_iterative(
        digraph, max_coi,
        root=coi_nodes, untouchable=untouchable,
        already_deleted=deleted
    )
    coi_size = len(coi)
    factor = float(coi_size) / float(max_coi)
    nothing_condition = (size0 <= max_coi)
    success_condition = (
        undershoot_tolerance <= factor and factor <= overshoot_tolerance
    )
    if nothing_condition:
        logger.debug(
            'Got a "null" COI reduction ({} -> {}/{})'
            .format(size0, coi_size, max_coi)
        )
    elif success_condition:
        logger.debug(
            'Got a "good" COI reduction ({} -> {}/{}).'.format(
                size0, coi_size, max_coi
            )
        )
    else:
        logger.debug(
            'Got a "bad" COI reduction ({} -> {}/{}).'
            .format(size0, coi_size, max_coi)
        )

    ## Convert from "nodes" back to "variables" and return.
    relevant_deleted = set.intersection(coi, deleted)
    return (
        {node2var[x] for x in relevant_deleted},
        coi_size, size0,
        {node2var[x] for x in coi}
    )



## Copyright (c) 2016, Blake C. Rawlings.
##
## This file is part of `st2smv`.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import copy
import json
import pprint

from . import ast
from . import plugins
from . import st
from . import utils

logger = utils.logging.getLogger(__name__)

## Convenience functions.
ONE_OR_MORE = 'ONE_OR_MORE'
ONE_AND_ONLY_ONE = 'ONE_AND_ONLY_ONE'
ONE_AT_MOST = 'ONE_AT_MOST'

## FSM-related nodes.
TRANS = 'TRANS'
DEFINE = 'DEFINE'
INVAR = 'INVAR'
CTRBL = 'CTRBL'

## Typed equality comparators.
EQ_BOOLEAN = '{}{}'.format(ast.EQ, ast.BOOLEAN)
NEQ_BOOLEAN = '{}{}'.format(ast.NEQ, ast.BOOLEAN)
binary_boolean_operators = (
    ast.binary_boolean_operators + (EQ_BOOLEAN, NEQ_BOOLEAN)
)
##
EQ_NUMERIC = '{}{}'.format(ast.EQ, ast.NUMERIC)
NEQ_NUMERIC = '{}{}'.format(ast.NEQ, ast.NUMERIC)
binary_numeric_operators = (
    ast.binary_arithmetic_operators + ast.inequality_comparators
    + (EQ_NUMERIC, NEQ_NUMERIC)
)


class ConversionError(Exception):
    """Error with converting part of the AST."""
    pass


class ConflictingTypesError(Exception):
    """Error with the types in the AST."""
    pass


class Var(object):
    """The intermediate representation of a variable."""

    def __init__(self, label, version, field=None):
        assert isinstance(label, utils.six.string_types)
        assert isinstance(version, int) or version is None
        assert isinstance(field, utils.six.string_types) or field is None
        self.label = label
        self.version = version
        self.field = field

    def __repr__(self):
        return 'ir.Var({}, {}, field={})'.format(
            self.label, self.version, self.field
        )

    def __str__(self):
        if self.field is None:
            ending = ''
        else:
            ending = '.{}'.format(self.field)
        if self.version is None:
            return self.label + ending
        else:
            return '{}_{}{}'.format(self.label, self.version, ending)

    def __lt__(self, other):
        return repr(self) < repr(other)

    def __eq__(self, other):
        if isinstance(other, Var):
            return (
                self.label == other.label
                and self.version == other.version
                and self.field == other.field
            )
        else:
            return NotImplemented

    def __hash__(self):
        return hash((self.label, self.version, self.field))

    def to_dict(self):
        return {
            'label': self.label,
            'version': self.version,
            'field': self.field,
        }

    def root_var(self):
        return Var(self.label, None, field=self.field)


class Converter(object):
    """Convert an AST to intermediate representation (IR)."""

    def __init__(self, loaded_plugins):
        ## `loaded_plugins` is a list of optional plugins to apply
        ## when building the model.
        self.loaded_plugins = loaded_plugins
        ## `plugin_options` is a dict that maps plugin option names to
        ## values.
        self.plugin_options = {}
        ## `assignment_count` stores the number of times each variable
        ## has had a value assigned to it.
        ##
        ## The keys are /AST/ variables.
        self.assignment_count = {}
        ## `connections` stores the number of times a variable's value
        ## depends on another variable, i.e.,
        ## `connections[var1][var2]` is the number of times `var2`
        ## appears directly in the assignment logic for `var`.
        ##
        ## The keys are /IR/ variables.
        self.connections = {}
        ## `referenced` is the set of variables that appear directly
        ## in the assignment logic for other variables.
        ##
        ## The members are /IR/ variables.
        self.referenced = set()
        ## `assigned` is the set of variables that are assigned a
        ## value somewhere in the program.
        ##
        ## The members are /AST/ variables.
        self.assigned = set()
        ## `defined` is the set of variables that have their value
        ## defined via some assignment logic in the model.
        ##
        ## The members are /IR/ variables.
        self.defined = set()
        ## `trivial` is the set of variables that only have
        ## "assignment logic" to enforce the PLC "input scan, evaluate
        ## logic, output scan" looping behavior.
        ##
        ## The members are /IR/ variables.
        self.trivial = set()
        ## `aliases` is the set of variables that are just aliases for
        ## other variables.
        ##
        ## The members are /IR/ variables.
        self.aliases = set()
        ## `ivars` is the set of input variables in the model.
        ##
        ## The members are /IR/ variables.
        self.ivars = set()
        ## `literals` is the set of literals that appear in the
        ## model.
        ##
        ## The members are /IR/ variables.
        self.literals = set()
        ## `comparisons` stores the comparison (equality, inequality,
        ## etc.) expressions that appear in the program.
        ##
        ## The members are tuples that include the AST and the /IR/
        ## variable.
        self.comparisons = []
        ## `numeric` and `boolean` expressions that are used as either
        ## numeric or boolean values.
        ##
        ## The members are AST nodes.
        self.numeric = set()
        self.boolean = set()
        ## `intermediate_representation` is the intermediate
        ## representation of the model; `logic_ir` is the
        ## assignment/transition logic.
        self.intermediate_representation = []
        self.logic_ir = []
        ## `special_variables` is a dict that tracks various types of
        ## "special" variables, so that
        ## `special_variables[variable_type]` is the set of special
        ## variables of type `variable_type` that appear in the model;
        ## these are plugin-specific.
        ##
        ## The members are /AST/ variables.
        self.special_variables = {}
        ## `context` is a list of boolean expressions that all must
        ## evaluate to "TRUE" for the inner code to be executed (e.g.,
        ## the conditions from a sequence of nested "IF" statements).
        ##
        ## The members are /IR/ nodes.
        self.context = []
        ## `if_depth` stores the number of nested "IF" statements that
        ## are currently in effect when processing an AST node.
        ##
        ## The value should be a non-negative integer.
        self.if_depth = 0
        ## `metadata_paths` is the list of file names that contain
        ## additional metadata related to the program.
        self.metadata_paths = []
        ## `sequence` is the sequence (if one was specified) of the
        ## program; this is plugin-specific.
        self.sequence = None
        ## `abstractions` stores the abstractions that will be aplpied
        ## to simplify the model.
        self.abstractions = {}
        ## `unconditional_assignments` is the set of variables that
        ## are assigned a value without checking any conditions.
        self.unconditional_assignments = set()
        ## `coi_info` is a dict that could be used to reconstruct the
        ## cone of influence information for the model.
        self.coi_info = {}

    def convert_ast(self, tree):
        """Convert the AST `tree`.

        This can be called multiple times, e.g. once for each
        statement in the AST.

        """
        if len(tree) == 0:
            return
        if isinstance(tree, list):
            if isinstance(tree[0], list):
                for statement in tree:
                    self.convert_statement(statement)
            else:
                self.convert_statement(tree)
        else:
            raise ast.AstError('Could not traverse the AST: {0}'.format(tree))

    def convert_statement(self, tree):
        """Convert the AST `tree` that corresponds to a single statement."""
        if tree[0] == ast.ASSIGN:
            self._convert_assignment(tree)
        elif tree[0] == ast.IF:
            self._convert_ifelse(tree)
        elif tree[0] in (ast.CALL, ast.FUNCTION_BLOCK_CALL):
            self._convert_instruction(tree)
        else:
            raise ConversionError(
                'Unrecognized statement: {0}'.format(tree)
            )

    def _convert_ifelse(self, tree):
        """Convert an IF...ELSE statement.

        This modifies `self`.

        """
        ## Check the structure of the tree and break it up.
        assert len(tree) == 2
        assert len(tree[1]) == 3
        cond = self.convert_expression(tree[1][0])
        if_branch = tree[1][1]
        else_branch = tree[1][2]

        ## Handle the IF branch.
        self.context.append(cond)
        self.if_depth += 1
        self.convert_ast(if_branch)

        ## Handle the ELSE branch.
        self.context[self.if_depth-1] = ['NOT', self.context[self.if_depth-1]]
        self.convert_ast(else_branch)
        self.context = self.context[:-1]
        self.if_depth -= 1

        ## Clean up the context.
        if self.if_depth == 0:
            assert len(self.context) == 0

    def _convert_assignment(self, tree):
        """Convert an assignment.

        This modifies `self`.

        """
        ## Get the right-hand-side expression.  If it isn't handled
        ## correctly, just mark the left-hand-side variable as being
        ## modified.
        try:
            rhs = self.convert_expression(tree[1][1])
        except ConversionError:
            logger.warning(
                'Could not convert the right-hand side, '
                'marking left-hand side as modified: {}'
                .format(tree)
            )
            self.convert_statement([ast.CALL, ['MODIFY', [tree[1][0]]]])
            return
        if self.if_depth > 0:
            ## This has to be before `lhs` is computed, because doing
            ## so increments the assignment count.
            lhs_old = self.convert_expression(tree[1][0], lhs=False)
        ## Add context/guards if necessary.
        if self.if_depth > 0:
            ## This assignment shows up under one or more IF/ELSE
            ## clauses, so it should be assigned a conditional value.
            assign_cond = self._conjoin(self.context)
            skip_cond = 'TRUE'
            rhs_with_context = [
                'case',
                [
                    [assign_cond, rhs],
                    [skip_cond, lhs_old]
                ]
            ]
            ## Get the left-hand-side expression.
            lhs = self.convert_expression(tree[1][0], lhs=True)
        else:
            ## Not part of any IF/ELSE block, i.e., an unconditional
            ## assignment.
            lhs = self.convert_expression(tree[1][0], lhs=True)
            self.unconditional_assignments.add(lhs)
            rhs_with_context = rhs
        ## Select how to decide the assignment type.
        ## Use the new `get_type` and `set_type` methods.
        branch_types = [
            self.get_type(lhs), self.get_type(rhs_with_context)
        ]
        if ast.BOOLEAN in branch_types:
            assert ast.NUMERIC not in branch_types
            assignment_type = ast.BOOLEAN
        elif ast.NUMERIC in branch_types:
            assert ast.BOOLEAN not in branch_types
            assignment_type = ast.NUMERIC
        else:
            assert all([x == ast.VALUE for x in branch_types])
            assignment_type = ast.VALUE
        if assignment_type in (ast.BOOLEAN, ast.NUMERIC):
            self.set_type(lhs, assignment_type)
            self.set_type(rhs_with_context, assignment_type)
        ## Create the assignment logic.
        self.defined.add(lhs)
        keyword = ast.ASSIGN
        if assignment_type == ast.BOOLEAN:
            self.logic_ir.append([keyword, [lhs, rhs_with_context]])
        elif assignment_type == ast.NUMERIC:
            self.logic_ir.append([keyword+'_NUMERIC', [lhs, rhs_with_context]])
        else:
            assert assignment_type == ast.VALUE
            self.logic_ir.append([keyword+'_VALUE', [lhs, rhs_with_context]])
        ## Keep track of which variables influence each other.
        variables = set.union(self.defined, self.referenced)
        rhs_symbols = get_variables(rhs_with_context)
        rhs_vars = {x for x in variables if x in rhs_symbols}
        for v in rhs_vars:
            self.mark_connection(lhs, v)

    def _convert_comment(self, tree):
        """Convert a comment."""
        raise NotImplementedError(
            'Comments should have been stripped from the AST.'
        )

    def _convert_instruction(self, tree):
        """Convert an instruction.

        This modifies `self`.

        """
        call_type = tree[0]
        procedure = tree[1][0]
        arguments = tree[1][1]
        if procedure == 'MODIFY':
            assert len(arguments) == 1
            utils.try_incr(self.assignment_count, Var(arguments[0], None), 1)
        else:
            logger.warning('Unrecognized instruction: {}'.format(tree))

    def convert_expression(self, tree, lhs=False):
        """Convert the AST `tree` that corresponds to an expression.

        This modifies `self` and returns the converted expression.

        """
        ## Base cases.
        if len(tree) == 0:
            return None
        elif isinstance(tree, utils.six.string_types):
            return self._convert_variable(tree, lhs=lhs)
        else:
            assert isinstance(tree, list)

        ## Check if any plugins handle the expression.
        active_plugins = plugins.actions.get_active_plugins(
            self.loaded_plugins, plugins.actions.SPECIAL_EXPRESSIONS
        )
        for plugin in active_plugins:
            if tree[0] in plugin.SPECIAL_EXPRESSION_TYPES:
                return plugin.convert_expression(self, tree, lhs=lhs)

        ## Left-hand side variables should have been handled by now.
        assert lhs is False

        ## Recursive cases.
        if len(tree) == 1:
            return [self.convert_expression(tree[0])]
        elif tree[0] in (ast.CALL, ast.FUNCTION_BLOCK_CALL):
            assert len(tree) == 2
            assert len(tree[1]) == 2
            return self.convert_expression(tree[1])
        elif tree[0] in ast.boolean_operators:
            return self._convert_boolean_expression(tree)
        elif tree[0] in ast.comparators:
            return self._convert_comparison(tree)
        elif tree[0] in (ast.arithmetic_operators + ('ABS', 'POS')):
            return self._convert_numeric_expression(tree)
        else:
            raise ConversionError(
                'Unrecognized expression: {0}'.format(tree)
            )

    def _convert_variable(self, tree, lhs=False, force_version=None):
        """Convert a variable (or a literal)."""
        assert isinstance(tree, utils.six.string_types)
        ## Literals.
        if tree == st.parse_symbol(tree, (st.literal_bool | st.literal_num)):
            return self._convert_literal(tree)
        ## Variables.
        if not lhs:
            utils.try_incr(self.assignment_count, Var(tree, None), 0)
        else:
            utils.try_incr(self.assignment_count, Var(tree, None), 1)
        ## Return the correct "version" of the variable.
        if force_version is None:
            version = self.assignment_count[Var(tree, None)]
        else:
            assert isinstance(force_version, int)
            version = force_version
        retval = Var(tree, version)
        ## Mark this variable's behavior/function.
        if not lhs:
            self.referenced.add(retval)
        else:
            self.assigned.add(Var(tree, None))
        ##
        return retval

    def _convert_literal(self, tree):
        """Convert a literal."""
        if tree == st.parse_symbol(tree, st.literal_bool):
            self.mark_boolean(tree)
            retval = Var(tree.upper(), None)
            self.set_type(retval, ast.BOOLEAN)
        else:
            assert tree == st.parse_symbol(tree, st.literal_num)
            self.mark_numeric(tree)
            retval = Var('{0}c'.format(tree.replace('.', 'p')), None)
            self.set_type(retval, ast.NUMERIC)
        self.literals.add(retval)
        return retval

    def _convert_boolean_expression(self, tree):
        """Convert a boolean expression."""
        assert tree[0] in ast.boolean_operators
        if tree[0] in ast.unary_boolean_operators:
            self.mark_boolean(tree[1])
            operand = self.convert_expression(tree[1])
            self.set_type(operand, ast.BOOLEAN)
            return [tree[0], operand]
        elif tree[0] in ast.binary_boolean_operators:
            self.mark_boolean(tree[1][0])
            self.mark_boolean(tree[1][1])
            operands = [self.convert_expression(x) for x in tree[1]]
            for operand in operands:
                self.set_type(operand, ast.BOOLEAN)
            return [
                tree[0],
                operands
            ]

    def _convert_comparison(self, tree):
        """Convert a numeric (in)equality comparison."""
        assert tree[0] in ast.comparators
        assert len(tree) == 2
        assert isinstance(tree[1], (tuple, list))
        assert len(tree[1]) == 2
        ##
        if tree[0] in ast.inequality_comparators:
            self.mark_numeric(tree[1][0])
            self.mark_numeric(tree[1][1])
        retval = [
            tree[0],
            [self.convert_expression(x) for x in tree[1]]
        ]
        ## Mark the comparison.
        self.mark_comparison(tree, retval)
        ##
        return retval

    def _convert_numeric_expression(self, tree):
        """Convert a numeric expression."""
        assert tree[0] in (ast.arithmetic_operators + ('ABS', 'POS'))
        assert len(tree) == 2
        ##
        if tree[0] in (ast.NEG, 'ABS', 'POS'):
            self.mark_numeric(tree[1])
            return [
                tree[0],
                self.convert_expression(tree[1])
            ]
        else:
            assert isinstance(tree[1], (tuple, list))
            assert len(tree[1]) == 2
            self.mark_numeric(tree[1][0])
            self.mark_numeric(tree[1][1])
            return [
                tree[0],
                [self.convert_expression(x) for x in tree[1]]
            ]

    def mark_connection(self, lhs, rhs):
        """Mark a connection, where `rhs` influences `lhs`."""
        assert isinstance(lhs, Var)
        assert isinstance(rhs, Var)
        if lhs not in self.connections:
            self.connections[lhs] = {}
        utils.try_incr(self.connections[lhs], rhs, 1)

    def mark_numeric(self, tree):
        """Try to mark the AST `tree` as numeric."""
        try:
            if tree in self.boolean:
                raise ConflictingTypesError(
                    'Tried to mark `{}` as numeric, '
                    'but it is already marked as Boolean.'
                    .format(tree)
                )
            self.numeric.add(tree)
        except TypeError:
            pass

    def mark_boolean(self, tree):
        """Try to mark `tree` as Boolean."""
        try:
            if tree in self.numeric:
                raise ConflictingTypesError(
                    'Tried to mark `{}` as Boolean, '
                    'but it is already marked as numeric.'
                    .format(tree)
                )
            self.boolean.add(tree)
        except TypeError:
            pass

    def is_boolean(self, tree):
        """Check if `tree` is marked as Bolean."""
        try:
            return str(tree) in self.boolean
        except TypeError:
            return False

    def is_numeric(self, tree):
        """Check if `tree` is marked as numeric."""
        try:
            return str(tree) in self.numeric
        except TypeError:
            return False

    def mark_comparison(self, ast_tree, ir_tree):
        """Mark the AST `tree` as a comparison.

        This actually stores a modified copy of `tree` that contains
        the current versions of the variables at this point in the
        code.  The translated IR `code` (variable name) is also
        stored along with the comparison to make adding constraints
        easier later.

        """
        substituted_ast_tree = copy.deepcopy(ast_tree)
        self._replace_variables(substituted_ast_tree)
        self.comparisons.append((substituted_ast_tree, ir_tree))

    def _replace_variables(self, tree):
        """Replace the values in `tree` with their current versions.

        N.B.: this modifies `tree` in place!  It does not affect any
        other variables though (i.e., `self` is not modified).

        """
        for i in range(len(tree)):
            if tree == st.parse_symbol(
                    tree, (st.literal_bool | st.literal_num)
            ):
                tree[i] = self._convert_literal(tree[i])
            elif tree[i] == st.parse_symbol(tree[i], st.identifier):
                try:
                    tree[i] = '{0}_{1}'.format(
                        tree[i],
                        self.assignment_count[Var(tree[i], None)]
                    )
                except KeyError:
                    ## Probably a keyword/function (e.g., "ABS").
                    pass
            elif isinstance(tree[i], list):
                self._replace_variables(tree[i])

    def _conjoin(self, expressions, top_level_call=True):
        """Returns the conjunction of a list of IR `expressions`."""
        expressions = copy.deepcopy(expressions)
        if top_level_call:
            for expression in expressions:
                self.set_type(expression, ast.BOOLEAN)
        if len(expressions) == 1:
            retval = (expressions[0])
        else:
            assert len(expressions) > 1
            retval = [
                ast.AND,
                [
                    (expressions[0]),
                    self._conjoin(expressions[1:], top_level_call=False)
                ]
            ]
        return retval

    def generate_model(self):
        """Produce the intermediate representation (IR) for the model.

        After all the statements in an AST have been processed, this
        method does some final processing and generates the output
        model (as IR).  The "final processing" includes defining
        aliases for some of the variables, enforcing the "input scan,
        evaluate logic, output scan" operation of a PLC, checking for
        infeasible readings/predicates, computing abstractions to
        simplify the output model, creating and writing
        process-independent tests (PITs) for the model, and more.

        """
        tree = []

        tree += self.logic_ir

        tree += self._generate_variable_aliases()

        ## Creating the transition logic may add transition
        ## restrictions to variables (the _0 versions of them) that
        ## were previously unrestricted; keep track of this.
        (newir, trivial) = self.generate_transition_logic()
        tree += newir
        self.trivial = set.union(self.trivial, trivial)

        ## Plugins.
        active_plugins = plugins.actions.get_active_plugins(
            self.loaded_plugins, plugins.actions.TRANSITION_LOGIC
        )
        for plugin in active_plugins:
            (newir, trivial) = plugin.generate_transition_logic(self)
            tree += newir
            self.trivial = set.union(self.trivial, trivial)
            logger.debug(
                'Added transition logic from the "{}" plugin.'
                .format(plugin.NAME)
            )

        ## Metadata.
        for mp in self.metadata_paths:
            newir = self.generate_metadata_constraints(mp)
            tree += newir

        ## This has to come after any modifications to `self.trivial`.
        tree = self._generate_variable_declarations() + tree

        ## Refinements (this may modify `self.connections`, so do it
        ## before computing abstractions).
        active_plugins = plugins.actions.get_active_plugins(
            self.loaded_plugins, plugins.actions.REFINEMENT
        )
        for plugin in active_plugins:
            (constraints, connections) = plugin.compute_refinements(self)
            for connection in connections:
                self.mark_connection(connection[0], connection[1])
                self.mark_connection(connection[1], connection[0])
            tree += constraints

        ## Abstractions.
        active_plugins = plugins.actions.get_active_plugins(
            self.loaded_plugins, plugins.actions.ABSTRACTION
        )
        if len(active_plugins) > 0:
            assert len(active_plugins) == 1
            if False:
                abstractions = active_plugins[0].compute_abstractions(self)
                assert abstractions[0] not in self.abstractions
                self.abstractions[abstractions[0]] = abstractions[1]
            ## Also save the COI info.
            self.coi_info = active_plugins[0].get_coi_info(self)

        ## Store the intermediate representation.
        self.intermediate_representation = tree

    def _generate_variable_declarations(self):
        """Generate "VAR" declarations as necessary.

        This modifies `self`, and returns the new variable
        declarations.

        """
        tree = []
        readings = sorted(
            self.referenced - self.literals - self.defined - self.trivial
        )
        assert all([isinstance(x, Var) for x in readings])
        for var in readings:
            ivar = Var('ivar_{}'.format(var.label), var.version)
            self.ivars.add(ivar)
            var_type = self.get_type(var)
            if var_type == ast.BOOLEAN:
                keyword = ast.ASSIGN
            elif var_type == ast.NUMERIC:
                keyword = ast.ASSIGN + '_' + ast.NUMERIC
            else:
                assert var_type == ast.VALUE
                keyword = ast.ASSIGN + '_' + ast.VALUE
            tree += [
                ['IVAR', ivar, var_type],
                ['VAR', var, var_type],
                [keyword, [['next', var], ivar]],
            ]
            self.mark_connection(var, ivar)
        return tree

    def _generate_variable_aliases(self):
        """Define some useful variable aliases.

        This modifies `self`, and returns the new aliases.

        """
        tree = []
        for var in sorted(self.assigned):
            assert isinstance(var, Var)
            assert var.version is None
            ## Define the base variable name to refer to the final
            ## modified value.
            if 'steps' in self.special_variables:
                if var in self.special_variables['steps']:
                    ## Special case, handled elsewhere.
                    continue
            if var == 'dt':
                ## Don't track the timer variables.
                continue
            version = Var(
                var.label, self.assignment_count[var], field=var.field
            )
            if self.is_boolean(var):
                ## Only Boolean variables "matter".
                tree.append([DEFINE, [var, version]])
            self.aliases.add(var)
            self.mark_connection(var, version)
        return tree

    def generate_transition_logic(self):
        """Enforce the variable transition logic.

        This modifies `self`, and returns the new transition logic.

        """
        tree = []
        trivial = set()
        for var in sorted(self.assigned):
            assert isinstance(var, Var)
            assert var.version is None
            ## Check if this variable should be handled by one of the
            ## plugins.
            skip = False
            for plugin in plugins.actions.get_active_plugins(
                    self.loaded_plugins, plugins.actions.TRANSITION_LOGIC
            ):
                if plugin.has_special_transition_logic(self, var):
                    skip = True
                    break
            if skip:
                continue
            ##
            var0 = Var(var.label, 0, field=var.field)
            if var0 in self.referenced:
                self.mark_connection(var0, var)
                if self.is_boolean(var):
                    tree.append([ast.ASSIGN, [['next', var0], var]])
                if var0 not in self.defined:
                    trivial.add(var0)
        return (tree, trivial)

    def generate_metadata_constraints(self, metadata_path):
        """Generate constraints on the model from some metadata.

        This reads a (JSON) file from `metadata_path` that defines
        some additional information about the system, and applies the
        corresponding constraints to the model.

        """
        ## Setup.
        tree = []
        if metadata_path == '':
            return tree
        ## Plugins.
        for plugin in plugins.actions.get_active_plugins(
                self.loaded_plugins, plugins.actions.METADATA_CONSTRAINTS
        ):
            new_ir = plugin.generate_metadata_constraints(
                self, metadata_path
            )
            tree += new_ir
        ##
        return tree

    def is_untouchable(self, variable):
        """Return whether or not `variable` is "untouchable".

        A variable is "untouchable" if it cannot (or should not) be
        abstracted away when simplifying the model.

        """
        return bool(variable in self.trivial)

    def get_type(self, tree):
        """Try to determine the type of the IR expression `tree`.

        This uses information from the nodes that appear in `tree`,
        along with variable type information that is already known.

        """
        if isinstance(tree, Var):
            if self.is_boolean(str(tree.root_var())):
                return ast.BOOLEAN
            elif self.is_numeric(str(tree.root_var())):
                return ast.NUMERIC
            else:
                return ast.VALUE
        elif isinstance(tree, (tuple, list)):
            assert len(tree) > 0
            if tree[0] in ast.arithmetic_operators + ('ABS', 'POS'):
                return ast.NUMERIC
            elif tree[0] in ast.boolean_operators + ast.comparators:
                return ast.BOOLEAN
            elif tree[0] == 'case':
                assert len(tree) == 2
                assert len(tree[1]) > 0
                case_types = [
                    self.get_type(x[1])
                    for x in tree[1]
                ]
                if ast.BOOLEAN in case_types:
                    assert ast.NUMERIC not in case_types
                    return ast.BOOLEAN
                elif ast.NUMERIC in case_types:
                    assert ast.BOOLEAN not in case_types
                    return ast.NUMERIC
                else:
                    assert all([x == ast.VALUE for x in case_types])
                    return ast.VALUE
                ## TODO: propagate types, make sure the conditionals
                ## are Boolean?
            else:
                print(tree)
                assert False
        else:
            print(tree)
            assert False

    def set_type(self, tree, new_type):
        """Try to set the type of the IR experssion `tree` to `new_type`.

        This moves through `tree` and calls itself recursively to set
        as many variables as possible (correctly) to `new_type`.

        """
        assert new_type in (ast.BOOLEAN, ast.NUMERIC, ast.VALUE)
        ##
        old_type = self.get_type(tree)
        if new_type == ast.VALUE:
            assert old_type == ast.VALUE
            return
        else:
            assert old_type in (new_type, ast.VALUE)
            if  new_type == ast.BOOLEAN:
                func = self.mark_boolean
            elif new_type == ast.NUMERIC:
                func = self.mark_numeric
            else:
                assert False
        ##
        if isinstance(tree, Var):
            func(str(tree.root_var()))
        elif isinstance(tree, (tuple, list)):
            assert len(tree) > 0
            unambiguous = (
                ast.inequality_comparators
                + ast.arithmetic_operators + ('ABS', 'POS')
                + ast.boolean_operators
            )
            if tree[0] in unambiguous:
                ## Nothing to do, the types should already be correct.
                return
            elif tree[0] in ast.equality_comparators:
                ## TODO: at least make sure the types agree?
                return
            elif tree[0] == 'case':
                assert len(tree) == 2
                assert len(tree[1]) > 0
                for x in tree[1]:
                    self.set_type(x[1], new_type)
                ## TODO: set all the conditionals to Boolean?
            else:
                print(tree)
                assert False
        else:
            print(tree)
            assert False

    def infer_types(self):
        """Apply type inference to add type information to the IR."""
        inferrers = []
        ## Assignments.
        inferrers.append(
            (
                lambda x: x == 'ASSIGN_VALUE',
                lambda _: 'ASSIGN',
                lambda _: 'ASSIGN_NUMERIC',
            )
        )
        ## Equality comparisons.
        inferrers.append(
            (
                lambda x: x in ast.equality_comparators,
                lambda x: '{}{}'.format(x, ast.BOOLEAN),
                lambda x: '{}{}'.format(x, ast.NUMERIC),
            )
        )

        ## Apply (and reapply) each of the `inferrers` until there's
        ## no change.
        hits = [
            utils.get_locations(self.logic_ir, rule=x[0])
            for x in inferrers
        ]
        n_untyped = sum([len(x) for x in hits])
        if n_untyped > 0:
            logger.debug(
                'Attempting to infer types for the ({}) untyped nodes.'
                .format(n_untyped)
            )
        else:
            return
        while True:
            ## Apply each inferrer.
            for inferrer in inferrers:
                self._add_types_to_untyped_nodes(
                    rule=inferrer[0],
                    on_bool=inferrer[1],
                    on_num=inferrer[2]
                )
            ## Find the remaining untyped nodes.
            new_hits = [
                utils.get_locations(self.logic_ir, rule=x[0])
                for x in inferrers
            ]
            ## If new type information was added, try again (the new
            ## information may propagate through the other
            ## `inferrers`).
            got_new_information = False
            for idx, locations in enumerate(new_hits):
                old_locations = hits[idx]
                if len(locations) != len(old_locations):
                    assert len(locations) < len(old_locations)
                    got_new_information = True
                    break
            if got_new_information:
                hits = new_hits
                continue
            else:
                break
        ##
        remaining_terms = []
        if any(len(x) > 0 for x in hits):
            for locations in hits:
                remaining_terms.extend(
                    utils.get_value_at_location(self.logic_ir, x[:-1])
                    for x in locations
                )
        n_rem = len(remaining_terms)
        if n_rem > 0:
            logger.warning(
                'Could not infer types for {} node(s):\n{}'
                .format(n_rem, pprint.pformat(remaining_terms))
            )
        elif n_untyped > 0:
            logger.debug(
                'Added types to all of the ({}) untyped nodes.'
                .format(n_untyped)
            )

    def _add_types_to_untyped_nodes(self, rule=None, on_bool=None, on_num=None):
        """Try to add type information to the "untyped" nodes.

        `rule` is the rule used to find the locations of untyped
        nodes, and `on_bool` and `on_num` are functions of a single
        argument which, when called with the operator of the untyped
        node as the argument, return the value that operator should be
        replaced with if it's found to be Boolean or numeric,
        respectively.

        """
        assert rule is not None
        assert on_bool is not None
        assert on_num is not None
        ##
        locations = utils.get_locations(
            self.logic_ir,
            rule=rule
        )
        n = len(locations)
        if n == 0:
            return
        new_typed = 0
        for location in locations:
            assert location[-1] == 0
            idx = location[:-1]
            tree = utils.get_value_at_location(self.logic_ir, idx)
            lhs_type = self.get_type(tree[1][0])
            rhs_type = self.get_type(tree[1][1])
            boolean = (lhs_type == ast.BOOLEAN) or (rhs_type == ast.BOOLEAN)
            numeric = (lhs_type == ast.NUMERIC) or (rhs_type == ast.NUMERIC)
            assert not (boolean and numeric)
            if boolean:
                self.set_type(tree[1][0], ast.BOOLEAN)
                self.set_type(tree[1][1], ast.BOOLEAN)
                utils.set_value_at_location(
                    self.logic_ir, location,
                    on_bool(tree[0])
                )
                new_typed += 1
            if numeric:
                self.set_type(tree[1][0], ast.NUMERIC)
                self.set_type(tree[1][1], ast.NUMERIC)
                utils.set_value_at_location(
                    self.logic_ir, location,
                    on_num(tree[0])
                )
                new_typed += 1
        if new_typed > 0:
            if n > 0:
                self._add_types_to_untyped_nodes(
                    rule=rule, on_bool=on_bool, on_num=on_num
                )

    def check_smv_variables(self):
        """Check for some expected behavior."""
        def check_vars(s):
            assert all([isinstance(x, Var) for x in s])
        check_vars(self.referenced)
        check_vars(self.defined)
        check_vars(self.trivial)
        check_vars(self.aliases)
        check_vars(self.ivars)
        check_vars(self.literals)
        for c in self.connections:
            assert isinstance(c, Var)
            check_vars(self.connections[c])
        for c in self.comparisons:
            assert isinstance(c[1], (Var, tuple, list))


class JSONEncoder(json.JSONEncoder):
    """A custom JSON encoder that can handle the `Var` class."""

    def default(self, obj):
        if isinstance(obj, Var):
            ## Encode `obj` such that it can be decoded "losslessly",
            ## back into a `Var` object.  See `json_decode_hook` for
            ## the decoding.
            return {str(type(obj)): obj.to_dict()}
        else:
            return json.JSONEncoder.default(self, obj)


def json_decode_hook(json_object):
    """JSON decoding hook to recover `Var` objects.

    Use this as the `object_hook` keyword argument of `json.load`.

    """
    if str(Var) in json_object:
        assert len(json_object) == 1
        members = json_object[str(Var)]
        return Var(members['label'], members['version'], field=members['field'])
    else:
        return json_object


def get_variables(tree):
    """Return the set of variables that appear in the IR `tree`."""
    locations = utils.get_locations(tree, rule=(lambda x: isinstance(x, Var)))
    terms = [utils.get_value_at_location(tree, l) for l in locations]
    retval = set(terms)
    return retval


def is_in_coi(tree, cone_of_influence):
    """Check if `tree` is in `cone_of_influence`."""
    locations = utils.get_locations(tree, rule=(lambda x: isinstance(x, Var)))
    terms = [utils.get_value_at_location(tree, l) for l in locations]
    variables = set(terms)
    if len(variables) == 0:
        return True
    retval = any([variable in cone_of_influence for variable in variables])
    return retval


def str2var(s, variables):
    """Try to match string `s` to one of the `Var`s in `variables`."""
    for variable in variables:
        if s == str(variable):
            return variable
    raise ValueError(s)



## Copyright (c) 2015-2016, Blake C. Rawlings.
##
## This file is part of `st2smv`.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import copy

import networkx

from ... import utils

logger = utils.logging.getLogger(__name__)


class AbstractionError(Exception):
    """An error when computing COI abstractions."""


def digraph_coi(dg, root):
    """Compute the cone of influence of `root`."""
    coi = set()
    for var in root:
        try:
            new_targets = networkx.algorithms.ancestors(dg, var)
        except networkx.NetworkXError:
            logger.warning('Unknown COI root node: {}'.format(var))
            new_targets = set()
        coi = set.union(coi, new_targets)
        coi.add(var)
    return coi


def ancestor_info(dg, targets, untouchable):
    """Get statistics about how the number of ancestors grows with search
    depth.

    """
    ## Convert to a standard form.
    if isinstance(targets, utils.six.string_types):
        targets = [targets]
    if untouchable is None:
        untouchable = []
    ## Get the shortest paths to the *set* of nodes, `target`.
    sp = {}
    for target in targets:
        sp_part = networkx.algorithms.shortest_path_length(
            dg, target=target
        )
        for u in untouchable:
            if u in sp_part:
                sp_part.pop(u)
        for source in sp_part:
            if source not in sp:
                sp[source] = sp_part[source]
            elif sp_part[source] < sp[source]:
                sp[source] = sp_part[source]
    ## Check how many nodes are "pulled in" as the path length
    ## increases.
    sp_ = copy.deepcopy(sp)
    depths = []
    i = 0
    while len(sp_) > 0:
        hit = []
        for key in sp_:
            if sp_[key] <= i:
                hit.append(key)
        depths.append(len(hit))
        for h in hit:
            sp_.pop(h)
        i += 1
    depths.append(len(sp_))
    ##
    return (sp, depths)



## Copyright (c) 2016-2017, Blake C. Rawlings.

import os
import setuptools

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
    readme_text = f.read()

setuptools.setup(
    name='st2smv',
    version='0.1.0',
    author='Blake C. Rawlings',
    author_email='blakecraw@gmail.com',
    description=(
        'A tool to convert Structured Text PLC code to an SMV model.'
    ),
    license='GPLv3',
    url='https://pypi.python.org/pypi/st2smv',
    packages=[
        'st2smv',
        'st2smv.plugins',
        'st2smv.plugins.connectivity',
        'st2smv.plugins.irrelevant_logic',
        'st2smv.plugins.stdlib',
        'st2smv.plugins.varlock',
    ],
    entry_points={
        'console_scripts': ['st2smv=st2smv.__main__:main']
    },
    long_description=readme_text,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
    ],
    install_requires=['networkx', 'pyparsing', 'six'],
    package_data={
        'st2smv': ['Makefile.run', '*.smv'],
    },
)



## Copyright (c) 2016, Blake C. Rawlings.
##
## This file is part of `st2smv`.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import json

from ... import ast
from ... import plugins
from ... import utils

logger = utils.logging.getLogger(__name__)

## Plugin information.
NAME = 'varlock'
ACTIONS = (
    plugins.actions.PROCESS_INDEPENDENT_TESTS,
    plugins.actions.POSTPROCESS,
)
OPTIONS = {
    'varlock_coi': (
        bool,
        (
            'when creating variable lock tests, '
            'include all the influencing variables'
        )
    )
}


def generate_pits(converter):
    pit_tmp = converter.assigned | converter.defined
    pit_tmp -= converter.aliases
    pit_tmp -= converter.ivars
    pit_tmp &= converter.unconditional_assignments
    pit_vars = sorted(
        {x for x in pit_tmp if (converter.get_type(x) == ast.BOOLEAN)}
    )

    pits = {'PIT_PREFIX': NAME}
    spec_types = (
        ('SPEC', 'varlock_potential'),
        ('SYNTH', 'varlock_guaranteed'),
    )
    include_coi = converter.plugin_options.get('varlock_coi')
    for var in pit_vars:
        ## Get the other variables that should factor in.
        coi = set()
        if not include_coi:
            others = set()
        else:
            coi_plugins = plugins.actions.get_active_plugins(
                converter.loaded_plugins, plugins.actions.INFLUENCE
            )
            for plugin in coi_plugins:
                coi |= plugin.get_cone_of_influence(converter, var)
            coi &= set(pit_vars)
            ## Ignore "known variable locks".
            known_varlocks_smv = _get_known_varlocks(converter)
            known_varlocks = {x for x in coi if str(x) in known_varlocks_smv}
            if var in known_varlocks:
                others = set()
            else:
                if len(known_varlocks) > 0:
                    logger.debug(
                        'Leaving known variable locks out of varlock test '
                        'for variable {}: {}'
                        .format(var, (known_varlocks & coi))
                    )
                others = (coi - known_varlocks) - {var}
        ## Build the specifications.
        inner = ' & '.join(
            [
                '(EF({0}) & EF(!{0}))'.format(x)
                for x
                in [var] + sorted(others)
            ]
        )
        pits[var] = []
        for spec_type in spec_types:
            pits[var].append(
                '{} NAME {}_{} := AG({});'
                .format(spec_type[0], spec_type[1], str(var), inner)
            )

    return pits


def _get_known_varlocks(converter):
    varlocks = set()
    for mp in converter.metadata_paths:
        with open(mp) as f:
            data = json.load(f)
        new_varlocks = set(data.get('varlocks', []))
        varlocks |= new_varlocks
    return varlocks


def get_postprocessing_categories():
    return (
        ('Potential Variable Locks', 'varlock_potential', 'false'),
        ('Guaranteed Variable Locks', 'varlock_guaranteed', 'false'),
    )



## Copyright (c) 2015-2017, Blake C. Rawlings.
##
## This file is part of `st2smv`.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import copy
import json
import networkx
import os
import shutil

import pkg_resources

import pyparsing

from . import ast
from . import ir
from . import plugins
from . import st
from . import utils

logger = utils.logging.getLogger(__name__)

NUSMV_2_6_0_KEYWORDS = (
    'MODULE', 'DEFINE', 'MDEFINE', 'CONSTANTS', 'VAR',
    'IVAR', 'FROZENVAR', 'INIT', 'TRANS', 'INVAR', 'SPEC',
    'CTLSPEC', 'LTLSPEC', 'PSLSPEC', 'COMPUTE', 'NAME',
    'INVARSPEC', 'FAIRNESS', 'JUSTICE', 'COMPASSION', 'ISA',
    'ASSIGN', 'CONSTRAINT', 'SIMPWFF', 'CTLWFF', 'LTLWFF',
    'PSLWFF', 'COMPWFF', 'IN', 'MIN', 'MAX', 'MIRROR', 'PRED',
    'PREDICATES', 'process', 'array', 'of', 'boolean', 'integer',
    'real', 'word', 'word1', 'bool', 'signed', 'unsigned',
    'extend', 'resize', 'sizeof', 'uwconst', 'swconst', 'EX',
    'AX', 'EF', 'AF', 'EG', 'AG', 'E', 'F', 'O', 'G', 'H', 'X',
    'Y', 'Z', 'A', 'U', 'S', 'V', 'T', 'BU', 'EBF', 'ABF', 'EBG',
    'ABG', 'case', 'esac', 'mod', 'next', 'init', 'union', 'in',
    'xor', 'xnor', 'self', 'TRUE', 'FALSE', 'count'
)
SYNTHSMV_KEYWORDS = ('SYNTH', 'CTRBL')
KEYWORDS = NUSMV_2_6_0_KEYWORDS + SYNTHSMV_KEYWORDS

## `_COMPARISONS` tracks Boolean variables that are created to take
## the place of numeric comparisons when converting IR to an SMV
## model.
##
## TODO: get rid of this global variable (hack).
_COMPARISONS = set()


def convert_symbol(symbol):
    """Convert a symbol from AST/IR syntax to SMV syntax."""
    mapping = {
        ast.AND: '&',
        ast.OR: '|',
        ast.XOR: 'xor',
        ast.NOT: '!',
        ast.LT: 'lt',
        ast.LEQ: 'le',
        ast.GT: 'gt',
        ast.GEQ: 'ge',
        ast.MLT: 'mul',
        ast.SUB: 'min',
        ast.ADD: 'pls',
        ast.DIV: 'div',
        ast.NEG: 'neg',
        'ABS': 'ABS',
        'POS': 'pos',
        ir.EQ_BOOLEAN: '=',
        ir.NEQ_BOOLEAN: '!=',
        ir.EQ_NUMERIC: 'eq',
        ir.NEQ_NUMERIC: 'ne',
        ast.EQ: 'eq',
        ast.NEQ: 'ne',
        ast.BOOLEAN: 'boolean',
    }
    for key in mapping:
        if key == symbol:
            return mapping[key]
    raise ir.ConversionError('Unrecognized symbol: {0}'.format(symbol))


@utils.timing(log_fun=logger.info, log_text='Converted the ST file: {:.2f}s')
def st_to_smv(
        fname,
        metadata_paths=(),
        directory=None,
        loaded_plugins=None,
        plugin_options=None,
):
    """Convert the Structured Text file `fname` to an SMV model.

    This writes the model (and possibly other output) and returns the
    `Converter` object.

    """
    if loaded_plugins is None:
        loaded_plugins = []
    if plugin_options is None:
        plugin_options = {}

    ## Parse the input file.
    tree = st.get_ast(fname)

    ## Check for any plugins that do AST transformations, and apply
    ## them.
    for plugin in plugins.actions.get_active_plugins(
            loaded_plugins, plugins.actions.AST_TRANSORMATIONS
    ):
        transformer = plugin.Transformer()
        transformed = transformer.transform_ast(tree)
        tree = transformed
        logger.debug(
            'Applied AST transformations from the "{}" plugin.'
            .format(plugin.NAME)
        )

    ## Save the AST in `directory`.
    ast.write_ast(tree, directory=directory)

    ## Convert the AST to IR.
    conv = ir.Converter(loaded_plugins)
    conv.metadata_paths = metadata_paths
    conv.plugin_options = plugin_options
    conv.convert_ast(tree)
    conv.infer_types()
    conv.generate_model()

    ## Make the IR an SMV module.
    initializing = ir.Var('initializing', None)
    conv.intermediate_representation = [
        ['MODULE', 'main'],
        [ast.ASSIGN, [['init', initializing], ir.Var('TRUE', None)]],
        [ast.ASSIGN, [['next', initializing], ir.Var('FALSE', None)]],
    ] + conv.intermediate_representation

    ## Run some sanity checks on the IR.
    conv.check_smv_variables()

    ## Write some output now.
    if directory is not None:
        if not os.path.isdir(directory):
            os.mkdir(directory)
    write_tests(conv, directory)
    if len(conv.abstractions) > 0:
        write_abstractions(conv, directory)
    if len(conv.coi_info) > 0:
        write_coi_info(conv, directory)
    copy_auxiliary(conv, directory)

    ## Write the output.
    path = directory + os.sep + 'model.json'
    with open(path, 'w') as f:
        json.dump(
            conv.intermediate_representation, f,
            indent=2, separators=(',', ': '),
            sort_keys=True,
            cls=ir.JSONEncoder
        )
        f.write('\n')

    return conv


def write_tests(conv, directory):
    """Write process-independent tests (PITs) in `directory`/pits/."""

    ## Plugins.
    plugin_pits = []
    for plugin in plugins.actions.get_active_plugins(
            conv.loaded_plugins, plugins.actions.PROCESS_INDEPENDENT_TESTS
    ):
        new = plugin.generate_pits(conv)
        for old in plugin_pits:
            assert old['PIT_PREFIX'] != new['PIT_PREFIX']
        plugin_pits.append(new)

    ## Combine the tests.
    tests = {}
    pit_dicts = plugin_pits
    for pit_dict in pit_dicts:
        for key in pit_dict:
            if key == 'PIT_PREFIX':
                continue
            if key not in tests:
                tests[key] = []
            tests[key] += pit_dict[key]

    ## Write the PITs.
    pit_dir = directory + os.sep + 'pits'
    if not os.path.isdir(pit_dir):
        os.mkdir(pit_dir)
    for key in tests:
        path = pit_dir + os.sep + '{}.smv'.format(key)
        with open(path, 'w') as f:
            f.write('\n'.join(tests[key]))
            f.write('\n')

    return


def write_abstractions(conv, directory):
    """Write the abstractions in `directory`/abstractions."""
    abstractions_dir = directory + os.sep + 'abstractions'
    if not os.path.isdir(abstractions_dir):
        os.mkdir(abstractions_dir)
    for abstraction_type in conv.abstractions:
        for key in conv.abstractions[abstraction_type]:
            path = (
                abstractions_dir
                + os.sep
                + '{}-{}.json'.format(abstraction_type, key)
            )
            with open(path, 'w') as f:
                json.dump(
                    conv.abstractions[abstraction_type][key], f,
                    indent=2, separators=(',', ': '),
                    sort_keys=True,
                    cls=ir.JSONEncoder
                )
                f.write('\n')
    return


def write_coi_info(conv, directory):
    """Write the COI info in `directory`."""
    with open(directory + os.sep + 'coi_info.json', 'w') as f:
        json.dump(
            conv.coi_info, f,
            indent=2, separators=(',', ': '),
            sort_keys=True,
            cls=ir.JSONEncoder
        )
        f.write('\n')
    return


def copy_auxiliary(conv, directory):
    """Copy some auxiliary files (SMV modules, etc.) to `directory`."""
    aux_names = ['Makefile.run']
    if 'timers' in conv.special_variables:
        aux_names.append('delay_timer.smv')
    for aux_name in aux_names:
        aux_location = pkg_resources.resource_filename(__name__, aux_name)
        shutil.copyfile(
            aux_location,
            directory + os.sep + aux_name
        )


def combine(infiles, metadatas, variables, loaded_plugins, plugin_options):
    """Combine multiple SMV components to form a single model.

    The input files `infiles` can be a mix of either JSON files
    containing IR or raw SMV files, and the metadata files `metadata`
    should be JSON files with additional metadata, such as
    abstractions (which are applied when combining the files).

    """
    _COMPARISONS.clear()
    (cone_of_influence, deleted) = get_coi_and_deleted_variables(
        metadatas, variables, loaded_plugins, plugin_options
    )
    modules = []
    raw = []
    output = []
    for infile in infiles:
        with open(infile, 'r') as f:
            if infile.endswith('smv'):
                lines = f.readlines()
                if any(['MODULE' in l for l in lines]):
                    modules += lines
                else:
                    raw += lines
            else:
                ## Structured (JSON) input, easy to do the
                ## replacements.
                assert infile.endswith('json')
                statements = json.load(f, object_hook=ir.json_decode_hook)
                intermediate_representation = []
                for statement in statements:
                    if variables is None:
                        pass
                    else:
                        ## Skip assignments to non-influencing variables.
                        if statement[0] in (ast.ASSIGN, ir.DEFINE):
                            ## Find out what the LHS variable is.
                            if isinstance(statement[1][0], ir.Var):
                                var = statement[1][0]
                            else:
                                assert statement[1][0][0] in ('init', 'next')
                                var = statement[1][0][1]
                            if not isinstance(var, ir.Var):
                                logger.error(
                                    'Could not determine LHS variable '
                                    'in assignment: {}'
                                    .format(statement)
                                )
                                assert False
                            ## Skip any variables that aren't in the cone
                            ## of influence.
                            if var not in cone_of_influence:
                                if var in deleted:
                                    logger.error(
                                        'Deleted variable not in COI: {}'
                                        .format(var)
                                    )
                                    assert False
                                if var.field is None:
                                    continue
                            ## Replace the `deleted` variables with dummy
                            ## inputs.
                            if var in deleted:
                                if var not in cone_of_influence:
                                    logger.error(
                                        'Deleted variable not in COI: {}'
                                        .format(var)
                                    )
                                    assert False
                                ivar = 'ivar_{}'.format(var)
                                intermediate_representation.append(
                                    ('IVAR', ivar, 'boolean')
                                )
                                intermediate_representation.append(
                                    ('VAR', var, 'boolean')
                                )
                                intermediate_representation.append(
                                    (ast.ASSIGN, (('next', var), ivar))
                                )
                                continue
                        ## Skip non-influencing statements.
                        elif not ir.is_in_coi(statement, cone_of_influence):
                            continue
                    ## At this point, we either are not doing COI
                    ## reduction, or we can't prove that the statement
                    ## doesn't influence the specification, so process
                    ## the statement.
                    intermediate_representation.append(statement)
                output += ir_to_lines(intermediate_representation)

    if len(_COMPARISONS) > 0:
        output.append('-- Numeric comparisons, as Boolean variables:')
        for comparison in sorted(_COMPARISONS):
            output.extend(
                [
                    'IVAR ivar_{} : boolean;'.format(comparison),
                    'VAR {} : boolean;'.format(comparison),
                    'TRANS next({0}) = ivar_{0};'.format(comparison),
                ]
            )

    for line in modules + output + raw:
        print(line.strip())


def get_coi_and_deleted_variables(
        metadatas, variable_strings, loaded_plugins, plugin_options
):
    """Compute the cone of influence and deleted variables."""
    initializing = ir.Var('initializing', None)
    deleted_variables = set()
    cone_of_influence = {initializing}
    if variable_strings is None:
        ## Load the already-generated COI info.
        for metadata in metadatas:
            if os.path.exists(metadata):
                ## Check the metadata file for things like deleted nodes,
                ## etc.
                with open(metadata, 'r') as f:
                    s = f.read()
                    try:
                        j = json.loads(s, object_hook=ir.json_decode_hook)
                    except ValueError:
                        logger.warning(
                            'Could not read JSON metadata: {}'.format(metadata)
                        )
                        j = {}
                    deleted_variables |= set(j.get('deleted', []))
                    cone_of_influence |= set(j.get('coi', []))
            else:
                if len(metadata) > 0:
                    logger.warning('Missing metadata file: {}'.format(metadata))
    else:
        ## Compute the COI info on the fly.
        assert len(metadatas) == 1
        with open(metadatas[0], 'r') as f:
            s = f.read()
            j = json.loads(s, object_hook=ir.json_decode_hook)
        edges = j['edges']
        node2var = {int(key): value for (key, value) in j['node2var'].items()}
        untouchable = j['untouchable']
        digraph = networkx.DiGraph(edges)
        all_variables = {node2var[k] for k in node2var}
        variables = {ir.str2var(x, all_variables) for x in variable_strings}
        active_plugins = plugins.actions.get_active_plugins(
            loaded_plugins, plugins.actions.ABSTRACTION
        )
        assert len(active_plugins) == 1
        plugin = active_plugins[0]
        (
            deleted, size, size0, target, coi
        ) = plugin.get_variable_cuts(
            digraph, node2var,
            variables, plugin_options, untouchable,
        )
        cone_of_influence |= coi
        deleted_variables |= deleted
    ## Do some cleanup and reporting.
    structures = set()
    for coi_var in cone_of_influence:
        if coi_var.field is not None:
            structure = copy.deepcopy(coi_var)
            structure.field = None
            structures.add(structure)
    cone_of_influence |= structures
    for var in deleted_variables:
        logger.debug('Deleting variable logic: {}'.format(var))
        assert isinstance(var, ir.Var)
    ##
    return (cone_of_influence, deleted_variables)


def ir_to_lines(tree, insert_comments=False):
    """Convert the IR `tree` to lines of SMV code."""
    assert isinstance(tree, (tuple, list))
    VARs = set()
    IVARs = set()
    lines = []
    for chunk in tree:
        assert isinstance(chunk, (tuple, list))
        if chunk[0] == 'MODULE':
            assert len(chunk) == 2
            lines.append(
                '{} {}'.format(
                    chunk[0],
                    ir_to_expression(chunk[1])
                )
            )
        elif chunk[0] in ('VAR', 'IVAR'):
            assert len(chunk) == 3
            var = chunk[1]
            rhs = chunk[2]
            ## Track declarations.
            if chunk[0] == 'VAR':
                category = VARs
            elif chunk[0] == 'IVAR':
                category = IVARs
            else:
                assert False
            if var not in category:
                category.add(var)
                if rhs not in (ast.NUMERIC, ast.VALUE):
                    lines.append(
                        '{} {} : {};'.format(
                            chunk[0],
                            ir_to_expression(var),
                            ir_to_expression(rhs)
                        )
                    )
                else:
                    logger.debug(
                        'Skipping variable declaration: {}'.format(chunk)
                    )
        elif chunk[0] == ast.ASSIGN:
            assert len(chunk) == 2
            assert len(chunk[1]) == 2
            lhs = ir_to_expression(chunk[1][0])
            rhs = ir_to_expression(chunk[1][1])
            ## Track declarations.
            if isinstance(chunk[1][0], (tuple, list)):
                assert len(chunk[1][0]) == 2
                assert chunk[1][0][0] in ('next', 'init')
                lhs_var = chunk[1][0][1]
            else:
                lhs_var = chunk[1][0]
                assert isinstance(lhs_var, ir.Var)
            if not (isinstance(lhs_var, ir.Var) or lhs_var is None):
                print(lhs_var)
                print(type(lhs_var))
                assert False
            if isinstance(lhs_var, ir.Var):
                if lhs_var not in VARs:
                    VARs.add(lhs_var)
                    lines.append('VAR {} : {};'.format(lhs_var, 'boolean'))
            ##
            lines.append('ASSIGN {} := {};'.format(lhs, rhs))
        elif chunk[0] in (ir.CTRBL, ir.INVAR):
            assert len(chunk) == 2
            lines.append(
                '{} {};'.format(
                    chunk[0],
                    ir_to_expression(chunk[1])
                )
            )
        elif chunk[0] == ir.TRANS:
            assert len(chunk) == 2
            lines.append(
                '{} {};'.format(
                    'TRANS',
                    ir_to_expression(chunk[1])
                )
            )
        elif chunk[0] == ir.DEFINE:
            assert len(chunk) == 2
            assert len(chunk[1]) == 2
            lines.append(
                '{} {} := {};'.format(
                    chunk[0],
                    ir_to_expression(chunk[1][0]),
                    ir_to_expression(chunk[1][1])
                )
            )
        elif chunk[0] == 'PIT':
            assert len(chunk) == 2
            lines.append('\n'.join(chunk[1]))
        elif chunk[0] == 'ASSIGN_NUMERIC':
            if insert_comments is False:
                pass
            else:
                lines.append(
                    '-- numeric assignment to {}'
                    .format(ir_to_expression(chunk[1][0]))
                )
        elif chunk[0] == 'ASSIGN_VALUE':
            if insert_comments is False:
                pass
            else:
                lines.append(
                    '-- untyped assignment to {}'
                    .format(ir_to_expression(chunk[1][0]))
                )
        else:
            logger.error(
                'Unrecognized statement in intermediate representation: {}'
                .format(repr(chunk))
            )
    assert len(set.intersection(VARs, IVARs)) == 0
    return lines


def ir_to_expression(tree, no_grouping=False):
    """Convert the IR expression `tree` to an SMV expression."""
    ## Base cases.
    if isinstance(tree, (ir.Var, utils.six.string_types)):
        try:
            return convert_symbol(tree)
        except ir.ConversionError:
            return str(tree)
    else:
        assert isinstance(tree, (tuple, list))
        assert len(tree) > 0

    ## Recursive cases.
    if len(tree) == 1:
        if no_grouping:
            format_string = '{}'
        else:
            format_string = '({})'
        return format_string.format(
            ir_to_expression(tree[0], no_grouping=no_grouping)
        )
    elif tree[0] in ('next', 'init'):
        assert len(tree) == 2
        return '{}({})'.format(tree[0], ir_to_expression(tree[1]))
    elif tree[0] in ast.unary_boolean_operators:
        assert len(tree) == 2
        return '({} {})'.format(
            convert_symbol(tree[0]), ir_to_expression(tree[1])
        )
    elif tree[0] in ir.binary_boolean_operators:
        assert len(tree) == 2
        return '({} {} {})'.format(
            ir_to_expression(tree[1][0]),
            convert_symbol(tree[0]),
            ir_to_expression(tree[1][1]),
        )
    elif tree[0] in ast.unary_arithmetic_operators + ('ABS', 'POS'):
        assert len(tree) == 2
        rhs = ir_to_expression(tree[1], no_grouping=True)
        return ir.Var(
            '{0}__{1}'.format(convert_symbol(tree[0]), rhs),
            None
        )
    elif tree[0] in ir.binary_numeric_operators + ast.comparators:
        assert len(tree) == 2
        retval = ir.Var(
            '{0}__{1}__{2}'.format(
                ir_to_expression(tree[1][0], no_grouping=True),
                convert_symbol(tree[0]),
                ir_to_expression(tree[1][1], no_grouping=True),
            ), None
        )
        if tree[0] not in ast.binary_arithmetic_operators:
            _COMPARISONS.add(retval)
        return retval
    elif tree[0] == 'case':
        assert len(tree) == 2
        cases = [
            '{} : {};'.format(
                ir_to_expression(pair[0]),
                ir_to_expression(pair[1]),
            )
            for pair in tree[1]
        ]
        return 'case {} esac'.format(
            ' '.join(cases)
        )
    elif tree[0] == ir.ONE_OR_MORE:
        assert len(tree[1]) > 0
        return '(count({}) > {})'.format(
            ', '.join([str(ir_to_expression(x)) for x in tree[1]]),
            ir.Var('0', None)
        )
    elif tree[0] == ir.ONE_AT_MOST:
        assert len(tree[1]) > 0
        return '(count({}) <= {})'.format(
            ', '.join([str(ir_to_expression(x)) for x in tree[1]]),
            ir.Var('1', None)
        )
    elif tree[0] == ir.ONE_AND_ONLY_ONE:
        assert len(tree[1]) > 0
        return '(count({}) = {})'.format(
            ', '.join([str(ir_to_expression(x)) for x in tree[1]]),
            ir.Var('1', None)
        )
    elif len(tree) > 1:
        logger.debug(
            'Treating the TREE as a function call: {}'.format(repr(tree))
        )
        return '{}({})'.format(
            ir_to_expression(tree[0]),
            ', '.join(
                [str(ir_to_expression(x)) for x in tree[1:]]
            )
        )
    else:
        logger.error(
            'Unrecognized expression in intermediate representation: {}'
            .format(repr(tree))
        )


def postprocess(loaded_plugins):
    """Post-process the test results in the current directory."""
    results = {}
    times = {}

    ## Find the output files.
    files = os.listdir('.')
    cex_suffix = '.cex'
    prop_suffix = '.prop'
    cex_paths = []
    prop_paths = []
    for fname in files:
        if fname.endswith(cex_suffix):
            cex_paths.append(fname)
        elif fname.endswith(prop_suffix):
            prop_paths.append(fname)
    ## TODO: make sure that `cex` and `prop` match up
    cex_paths = sorted(cex_paths)
    prop_paths = sorted(prop_paths)

    ## Read the output.
    cexs = [_parse_counterexamples(x) for x in cex_paths]
    props = [_parse_properties(x) for x in prop_paths]
    timing_infos = [_parse_timing_info(x) for x in cex_paths]

    ## Apply any plugins that handle specific result types.
    active_plugins = plugins.actions.get_active_plugins(
        loaded_plugins, plugins.actions.POSTPROCESS
    )
    for idx, cex in enumerate(cexs):
        prop = props[idx]
        for plugin in active_plugins:
            categories = plugin.get_postprocessing_categories()
            plugin_results = _postprocess_worker(cex, prop, categories)
            _merge_dicts_of_sets(results, plugin_results)

    ## Get the timing information.
    for idx, cex_path in enumerate(cex_paths):
        timing_info = timing_infos[idx]
        name = cex_path[:-len(cex_suffix)]
        times[name] = timing_info
    with open('timing.json', 'w') as f:
        json.dump(
            times, f,
            indent=2, separators=(',', ': '), sort_keys=True
        )
        f.write('\n')

    ## Print the results.
    if False:
        for key in sorted(results):
            value = results[key]
            print('## {} ({})'.format(key, len(value)))
            for label in sorted(value):
                print(label)
    with open('summary.json', 'w') as f:
        results_serializable = {
            key: sorted(results[key])
            for key in sorted(results)
        }
        json.dump(
            results_serializable, f,
            indent=2, separators=(',', ': '), sort_keys=True
        )
        f.write('\n')


def _postprocess_worker(cex, prop, categories):
    results = {}
    for p in prop:
        property_specification = p[0][2].strip()
        property_logic = p[1][0]
        name = p[1][3]
        ## Find the matching result.
        for c in cex:
            counterexample_specification = c[1].strip()
            counterexample_logic = c[0]
            if counterexample_specification == property_specification:
                if property_logic == 'CTL':
                    if counterexample_logic[0] == 'specification':
                        break
                if property_logic == 'SYNTH':
                    if counterexample_logic[0] == '(SYNTH)':
                        break
        else:
            assert False
        outcome = c[-1]

        for category in categories:
            result_type = category[0]
            property_type = category[1]
            outcome_test = category[2]
            if result_type not in results:
                results[result_type] = set()
            if name.startswith(property_type):
                if outcome == outcome_test:
                    label = name[len(property_type)+1:]
                    results[result_type].add(label)

    return results


def _merge_dicts_of_sets(current, new):
    """Merge the elements of `new` into `current` (via set union).

    Both `new` and `current` must be `dict`s whose values are `set`s.
    This modifies `current`.

    """
    for key in new:
        if key in current:
            current[key] |= new[key]
        else:
            current[key] = new[key]


def _parse_counterexamples(path):
    """Parse the counterexample file `path`."""
    BEGIN = pyparsing.Literal('--')
    LPAR = pyparsing.Literal('(')
    RPAR = pyparsing.Literal(')')
    TYPE = pyparsing.Combine(LPAR + pyparsing.Word(pyparsing.alphas) + RPAR)
    SPECIFICATION = pyparsing.Literal('specification')
    IS = pyparsing.Literal('is')
    OUTCOME = (pyparsing.Literal('true') | pyparsing.Literal('false'))

    result = pyparsing.Group(
        BEGIN.suppress()
        + pyparsing.Group(pyparsing.Optional(TYPE) + SPECIFICATION)
        + pyparsing.SkipTo(IS + OUTCOME, include=True)
    )

    with open(path) as f:
        lines = f.readlines()
    matches = []
    for line in lines:
        if not line.startswith('--'):
            continue
        else:
            try:
                match = result.parseString(line)[0].asList()
            except pyparsing.ParseException:
                continue
            matches.append(match)
    return matches


def _parse_properties(path):
    """Parse the properties file `path`."""
    ## First line.
    NUMBER = pyparsing.Word(pyparsing.nums)
    COLON = pyparsing.Literal(':')
    line_1 = pyparsing.Group(NUMBER + COLON + pyparsing.restOfLine)
    ## Second line.
    LSQ = pyparsing.Literal('[')
    RSQ = pyparsing.Literal(']')
    STUFF = pyparsing.Regex(r'[\w/]+')
    line_2 = pyparsing.Group(
        LSQ.suppress() + STUFF + STUFF + STUFF + STUFF + RSQ.suppress()
    )
    ## Parse line-by-line.
    first_lines = []
    second_lines = []
    with open(path, 'r') as f:
        lines = f.readlines()
    for line in lines:
        stripped = line.strip()
        if stripped.startswith(('***', '---')) or len(stripped) == 0:
            continue
        try:
            parsed = line_1.parseString(line)
            first_lines.append(parsed.asList()[0])
            continue
        except pyparsing.ParseException:
            pass
        try:
            parsed = line_2.parseString(line)
            second_lines.append(parsed.asList()[0])
            continue
        except pyparsing.ParseException:
            pass
    assert len(first_lines) == len(second_lines)
    ##
    return list(zip(first_lines, second_lines))


def _parse_timing_info(path):
    """Read the timing information form the SMV output at `path`."""
    ## Define a parser for the timing info line.
    DIGITS = pyparsing.Word(pyparsing.nums)
    NUMBER = pyparsing.Combine(DIGITS + '.' + DIGITS)
    BEGINNING = 'User time'
    parser = BEGINNING + NUMBER + pyparsing.Word(pyparsing.alphas)
    ## Find the lines in the file with timing info.
    with open(path) as f:
        lines = f.readlines()
    relevant_lines = [line for line in lines if line.startswith(BEGINNING)]
    if not len(relevant_lines) == 1:
        logger.warning(path)
        assert False
    relevant_line = relevant_lines[0]
    ## Parse the timing info line.
    parsed = parser.parseString(relevant_line).asList()
    user_time = float(parsed[1])
    units = parsed[2]
    return (user_time, units)



## Copyright (c) 2016, Blake C. Rawlings.
##
## This file is part of `st2smv`.

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

import argparse
import cProfile
import importlib

from . import ast
from . import smv
from . import st
from . import utils

logger = utils.logging.getLogger(__name__)


def load_plugin(loaded_plugins, plugin_name):
    """Load the plugin `plugin_name`.  This modifies `loaded_plugins` in
    place.

    """
    try:
        plugin = importlib.import_module(
            '.plugins.{0}.{0}'.format(plugin_name), package=__package__
        )
        if plugin not in loaded_plugins:
            loaded_plugins.append(plugin)
            logger.debug(
                'Loaded the "{}" plugin.'.format(plugin.NAME)
            )
        else:
            logger.warning(
                'Tried to load the "{}" plugin multiple times, skipping.'
                .format(plugin.NAME)
            )
    except ImportError as e:
        logger.error(
            'Failed to load the "{}" plugin: {}.'.format(plugin_name, e)
        )


def run(profiler):

    argument_parser = argparse.ArgumentParser(
        description=(
            'Convert Structured Text (ST) code to an SMV model.'
        )
    )

    commands = argument_parser.add_mutually_exclusive_group(required=True)

    commands.add_argument(
        '--convert',
        default=False, action='store_true',
        help='convert an ST input file to an SMV model'
    )

    commands.add_argument(
        '--combine',
        default=False, action='store_true',
        help=(
            'combine one or more files '
            '(either raw SMV files, '
            'or JSON files containing an intermediate representation) '
            'into a single SMV model'
        )
    )

    commands.add_argument(
        '--parse',
        default=False, action='store_true',
        help='parse the ST input and print the abstract syntax tree (AST)'
    )

    commands.add_argument(
        '--postprocess',
        default=False, action='store_true',
        help='postprocess the results in the current directory'
    )

    argument_parser.add_argument(
        '--input', '-i',
        nargs='+', action='store',
        default=[],
        help=(
            'specify the input files'
        )
    )

    argument_parser.add_argument(
        '--output-directory',
        action='store', type=str,
        default=None,
        help=(
            'specify the directory in which to store the output'
        )
    )

    argument_parser.add_argument(
        '--plugins',
        nargs='+', action='store',
        default=[],
        help=(
            'specify a list of plugins to (attempt to) load'
        )
    )

    argument_parser.add_argument(
        '--plugin-options',
        nargs='+', action='store',
        default=[],
        help=(
            'specify some plugin-specific options '
            'as a list, separated by spaces; '
            'each option should have the form "name" or "name=value" '
            '(without quotes)'
        )
    )

    argument_parser.add_argument(
        '--metadata', '-m',
        nargs='+', action='store',
        default=[],
        help='specify separate metadata files (JSON)'
    )

    argument_parser.add_argument(
        '--variables',
        nargs='+', action='store',
        default=None,
        help='specify the variables that appear in the specification'
    )

    argument_parser.add_argument(
        '--verbosity', '-v',
        action='store',
        default='info',
        choices=('error', 'warn', 'info', 'debug'),
        help='set the verbosity level; debug is the highest'
    )

    argument_parser.add_argument(
        '--profile',
        action='store_true',
        default=False,
        help=(
            'print profiling/performance information for `st2smv` itself'
        )
    )

    ## Parse (and check) the command-line arguments.
    args = argument_parser.parse_args()
    if args.convert and args.output_directory is None:
        argument_parser.error(
            'the `--convert` command reauires that '
            '`--output-directory` is specified'
        )
    if (args.convert or args.parse) and len(args.input) != 1:
        argument_parser.error(
            'the `--convert` and `--parse` commands '
            'require exactly one (1) `--input` (got {})'
            .format(len(args.input))
        )

    ## Start the profiler, if necessary.
    if args.profile:
        profiler.enable()

    verbosity_map = {
        'error': utils.logging.ERROR,
        'warn': utils.logging.WARN,
        'info': utils.logging.INFO,
        'debug': utils.logging.DEBUG,
    }
    utils.logging.basicConfig(level=verbosity_map[args.verbosity])

    ## Set up the plugins.
    loaded_plugins = []
    for plugin_name in args.plugins:
        load_plugin(loaded_plugins, plugin_name)

    ## Set up the plugin options.
    plugin_options = {}
    for string in args.plugin_options:
        pieces = string.split('=')
        ## Make sure one of the loaded plugins handles this option.
        option = pieces[0]
        option_is_handled = False
        for loaded_plugin in loaded_plugins:
            if option in loaded_plugin.OPTIONS:
                option_is_handled = True
                (option_type, option_help) = loaded_plugin.OPTIONS[option]
            if option_is_handled:
                break
        if not option_is_handled:
            raise LookupError(
                'Unrecognized plugin option: {}'.format(string)
            )
        ## Store the option value.
        if len(pieces) == 1:
            assert option_type is bool
            plugin_options[option] = True
        else:
            assert len(pieces) == 2
            plugin_options[option] = option_type(pieces[1])

    ## Run the command.
    if args.convert:
        _ = smv.st_to_smv(
            args.input[0],
            metadata_paths=args.metadata,
            directory=args.output_directory,
            loaded_plugins=loaded_plugins,
            plugin_options=plugin_options,
        )
    if args.parse:
        ast.dump_ast(args.input[0], st, loaded_plugins)
    if args.combine:
        smv.combine(args.input, args.metadata, args.variables, loaded_plugins, plugin_options)
    if args.postprocess:
        smv.postprocess(loaded_plugins)


def main():
    profiler = cProfile.Profile()
    try:
        run(profiler)
    except KeyboardInterrupt as e:
        raise e
    finally:
        profiler.disable()
        if len(profiler.getstats()) > 0:
            profiler.dump_stats('st2smv.prof')


if __name__ == '__main__':
    main()



## Copyright (c) 2016, Blake C. Rawlings.
##
## This file is part of `st2smv`.

SPECIAL_EXPRESSIONS = object()
TRANSITION_LOGIC = object()
UNTOUCHABLE = object()
METADATA_CONSTRAINTS = object()
AST_TRANSORMATIONS = object()
ABSTRACTION = object()
REFINEMENT = object()
PROCESS_INDEPENDENT_TESTS = object()
INFLUENCE = object()
POSTPROCESS = object()

def get_active_plugins(loaded_plugins, action):
    """Return the list of loaded plugins that perform `action`."""
    return [
        plugin
        for plugin in loaded_plugins
        if action in plugin.ACTIONS
    ]
